import numpy as np

def pose_align(model_poses, est_poses):
    model_points = []
    est_points = []
    for image_id in est_poses.keys():
        model_points.append(model_poses[image_id][:3, -1])
        est_points.append(est_poses[image_id][:3, -1])

    model_points = np.asarray(model_points).transpose()
    est_points = np.asarray(est_points).transpose()

    rot, trans, trans_error = traj_align(est_points, model_points)
    T = np.eye(4)
    T[:3, :3] = rot
    T[:3, -1] = trans

    aligned_est_poses = {
        image_id:
            (T.dot(est_poses[image_id]) \
                if image_id in est_poses else model_poses[image_id].copy())
        for image_id in model_poses.keys()
    }

    return aligned_est_poses, T


def traj_align(model, data):
    """Align two trajectories using the method of Horn (closed-form).
    https://vision.in.tum.de/data/datasets/rgbd-dataset/tools

    Input:
    model -- first trajectory (3xn)
    data -- second trajectory (3xn)

    Output:
    rot -- rotation matrix (3x3)
    trans -- translation vector (3x1)
    trans_error -- translational error per point (1xn)

    """
    model_zerocentered = model - model.mean(1)[:, None]
    data_zerocentered = data - data.mean(1)[:, None]

    W = np.zeros( (3,3) )
    for column in range(model.shape[1]):
        W += np.outer(model_zerocentered[:,column],data_zerocentered[:,column])
    U,d,Vh = np.linalg.linalg.svd(W.transpose())
    S = np.matrix(np.identity( 3 ))
    if(np.linalg.det(U) * np.linalg.det(Vh)<0):
        S[2,2] = -1
    rot = U*S*Vh
    trans = data.mean(1) - rot.dot(model.mean(1))

    model_aligned = rot.dot(model) + trans.transpose()
    alignment_error = model_aligned - data

    trans_error = np.sqrt(np.sum(np.multiply(alignment_error,alignment_error),0)).A[0]

    return rot,trans,trans_error


# https://github.com/intel-isl/Open3D/issues/2206
def is_rotation_righthanded(R, verbose=False):
    assert(len(R.shape) == 2)
    assert(R.shape[0] == 3)
    assert(R.shape[1] == 3)

    res = True

    # Check that axis are right-handed
    if not np.allclose(R[2], np.cross(R[0], R[1])):
        if verbose:
            print("x cross y != z", np.cross(R[0], R[1]), " != ", R[2])
        res = False

    if not np.allclose(R[0], np.cross(R[1], R[2])):
        if verbose:
            print("y cross z != x", np.cross(R[1], R[2]), " != ", R[0])
        res = False

    if not np.allclose(R[1], np.cross(R[2], R[0])):
        if verbose:
            print("z cross x != y", np.cross(R[2], R[0]), " != ", R[1])
        res = False

    # Check that axis' are right-handed
    if not np.allclose(R[:, 2], np.cross(R[:, 0], R[:, 1])):
        if verbose:
            print("x' cross y' != z'", np.cross(R[:, 0], R[:, 1]), " != ", R[:, 2])
        res = False

    if not np.allclose(R[:, 0], np.cross(R[:, 1], R[:, 2])):
        if verbose:
            print("y' cross z' != x'", np.cross(R[:, 1], R[:, 2]), " != ", R[:, 0])
        res = False

    if not np.allclose(R[:, 1], np.cross(R[:, 2], R[:, 0])):
        if verbose:
            print("z' cross x' != y'", np.cross(R[:, 2], R[:, 0]), " != ", R[:, 1])
        res = False

    return res


# https://github.com/intel-isl/Open3D/issues/2206
def correct_rotation(R, verbose=False):
    assert(len(R.shape) == 2)
    assert(R.shape[0] == 3)
    assert(R.shape[1] == 3)

    R = R.copy()
    if not is_rotation_righthanded(R, verbose=verbose):
        R[:, 2] = R[:, 2] * -1
        R = R / np.linalg.det(R)

    return R
