import numpy as np
import math
from ..utils.logging import get_main_logger
from tqdm import tqdm
logger = get_main_logger()

class CollisionDetector():
    """Initialize collision detector

    Args:
        points(np.array): scene points.
        finger_width(float): the width of the finger.
        batch_size(int): number of grasp pose to calculate in a batch.
        empty_thresh(int): minimum number of points to be in the grasping space.
        collision_thresh(int): maximum number of points to be in the collision space.
    """
    def __init__(
            self,
            points,
            finger_width = 0.01,
            batch_size = 1000,
            empty_thresh = 1,
            collision_thresh = 0,
        ):
        self.points = points
        self.finger_width = finger_width
        self.batch_size = batch_size
        self.empty_thresh = empty_thresh
        self.collision_thresh = collision_thresh
        logger.debug('Initialize collision detector with {} point'.format(len(self.points)))
        logger.debug('Finger width:{}'.format(finger_width))
        logger.debug('Batch size:{}'.format(batch_size))
        logger.debug('Empty thresh:{}'.format(empty_thresh))
        logger.debug('Collision thresh:{}'.format(collision_thresh))

    def detect_collision(self, grasp_group, approach_dist, device = 'cpu'):
        """Divide the given grasp group into several batch and apply collision detection to each group.

        Args:
            grasp_group(GraspGroup): the group to detect.
            approach_dist(float): the approach distance.
            device(str): 'cpu' or 'cuda:gpuid'(e.g. 'cuda:0'). Torch needs to be installed to use gpu.

        Returns:
            np.array(bool): True for collision and False for no collision.
            np.array(bool): True for empty and False for not empty.
        """
        num_grasp = len(grasp_group)
        batch_num = math.ceil(num_grasp / self.batch_size)
        logger.debug('Collision dection with batch number:{}, device:{}'.format(batch_num, device))
        grasping_mask = np.zeros(shape = (num_grasp), dtype = bool)
        collision_mask = np.zeros(shape = (num_grasp), dtype = bool)
        for batch_id in tqdm(range(batch_num), 'Apply collision detection'):
            logger.debug('Collision detection on batch:{}'.format(batch_id))
            start = batch_id * self.batch_size
            end = min((batch_id + 1) * self.batch_size, num_grasp)
            batch_grasping_mask, batch_collision_mask = self._batch_detect_collision(
                grasp_group = grasp_group[start:end],
                approach_dist = approach_dist,
                device = device
            )
            grasping_mask[start:end] = batch_grasping_mask
            collision_mask[start:end] = batch_collision_mask
        return grasping_mask, collision_mask

    def _batch_detect_collision(self, grasp_group, approach_dist, device = 'cpu'):
        """Apply collision detection to given GraspGroup

        Args:
            grasp_group(GraspGroup): the group to detect.
            approach_dist(float): the approach distance.
            device(str): 'cpu' or 'cuda:gpuid'(e.g. 'cuda:0'). Torch needs to be installed to use gpu.

        Returns:
            np.array(bool): True for collision and False for no collision.
            np.array(bool): True for empty and False for not empty.
        """
        translations = grasp_group.translations
        rotations = grasp_group.rotation_matrices
        heights = grasp_group.heights[:, np.newaxis]
        depths = grasp_group.depths[:, np.newaxis]
        widths = grasp_group.widths[:, np.newaxis]
        if device == 'cpu':
            transformed_points = self.points[np.newaxis,:,:] - translations[:, np.newaxis,:]
            transformed_points = np.matmul(transformed_points, rotations)
            grasping_mask, collision_mask = self._get_collision_masks(
                transformed_points,
                heights,
                depths,
                widths,
                approach_dist
            )
        elif device.startswith('cuda'):
            try:
                import torch
                logger.debug('Torch is installed to use GPU for collision detection')
            except:
                logger.error('Torch needs to be installed to use GPU for collision detection')
                raise ValueError('Torch needs to be installed to use GPU for collision detection')
            transformed_points = torch.tensor(self.points[np.newaxis,:,:]).to(device) - torch.tensor(translations[:, np.newaxis,:]).to(device)
            rotations = torch.tensor(rotations).to(device)
            transformed_points = torch.matmul(transformed_points, rotations)
            grasping_mask, collision_mask = self._get_collision_masks(
                transformed_points,
                torch.tensor(heights).to(device),
                torch.tensor(depths).to(device),
                torch.tensor(widths).to(device),
                approach_dist
            )
            grasping_mask = grasping_mask.detach().cpu().numpy()
            collision_mask = collision_mask.detach().cpu().numpy()
        else:
            raise ValueError('Unknown device "{}" for collision detection'.format(device))
        grasping_mask = np.sum(grasping_mask, axis = 1) >= self.empty_thresh
        collision_mask = np.sum(collision_mask, axis = 1) <= self.collision_thresh
        return grasping_mask, collision_mask

    def _get_collision_masks(self, transformed_points, heights, depths, widths, approach_dist):
        """Calculate collision mask for given poses and points.

        Args:
            transformed_points(np.array or torch.tensor): The transformed points in gripper frame.
            heights(np.array or torch.tensor): The heights of the grippers.
            depths(np.array or torch.tensor): The depths of the grippers.
            widths(np.array or torch.tensor): The widths of the grippers.
            approach_dist(float): The approach distance.

        Returns:
            np.array or torch.tensor: The mask for points in grasping space and collision space.
        """
        # height mask
        mask_height = ((transformed_points[:,:,2] > -heights/2) & (transformed_points[:,:,2] < heights/2))

        # inner mask
        mask_inner_depth = ((transformed_points[:,:,0] > 0) & (transformed_points[:,:,0] < depths))
        mask_inner_width = ((transformed_points[:,:,1] > -widths / 2) & (transformed_points[:,:,1] < widths / 2))
        inner_mask = mask_height & mask_inner_depth & mask_inner_width

        # outer mask
        mask_outer_depth = ((transformed_points[:,:,0] > -approach_dist) & (transformed_points[:,:,0] < depths))
        mask_outer_width = ((transformed_points[:,:,1] > (-widths / 2 - self.finger_width)) & ((transformed_points[:,:,1] < widths / 2 + self.finger_width)))
        outer_mask = mask_height & mask_outer_depth & mask_outer_width
        collision_mask = outer_mask & (~inner_mask)
        grasping_mask = inner_mask
        return grasping_mask, collision_mask
