#!/usr/bin/env python

import argparse
from pathlib import Path
import itertools
import time
import subprocess
import multiprocessing
import logging
import shutil
import re
import copy
import tempfile
import tqdm

import numpy as np
from scipy.spatial.transform import Rotation
import open3d as o3d
import pyrender
import trimesh
import matplotlib.pyplot as plt
import cv2


def main(image_path, sparse_model):
    timestamp = str(int(time.time()))
    # timestamp = '1640646101'
    dense_model = image_path.parents[0] / 'dense_model_{}'.format(timestamp)
    dense_model.mkdir(exist_ok=True)

    cmd = [
        'colmap', 'image_undistorter',
        '--input_path', str(sparse_model),
        '--output_path', str(dense_model),
        '--image_path', str(image_path)
    ]
    ret = subprocess.call(cmd)
    if ret != 0:
        logging.warning('Problem with image undistorter, exiting.')
        exit(ret)

    cmd = [
        'colmap', 'patch_match_stereo',
        '--workspace_path', str(dense_model),
        '--PatchMatchStereo.filter', 'true',
        '--PatchMatchStereo.geom_consistency', 'false',
        '--PatchMatchStereo.depth_max', str(7),
        '--PatchMatchStereo.depth_min', str(0),
        '--PatchMatchStereo.gpu_index', str(0),
        '--PatchMatchStereo.allow_missing_files', str(1),
        '--PatchMatchStereo.max_image_size', str(640),
        '--PatchMatchStereo.cache_size', str(64),
    ]
    ret = subprocess.call(cmd)
    if ret != 0:
        logging.warning('Problem with patch match stereo, exiting.')
        exit(ret)

    cmd = [
        'colmap', 'stereo_fusion',
        '--workspace_path', str(dense_model),
        '--output_path', str(dense_model / 'fused.ply'),
    ]
    ret = subprocess.call(cmd)
    if ret != 0:
        logging.warning('Problem with stereo fusion, exiting.')
        exit(ret)

    # convert output to txt (TODO: work with binaries directly)
    new_sparse_model_bin = dense_model / 'sparse'
    new_sparse_model_txt = dense_model / 'sparse_txt'
    new_sparse_model_txt.mkdir(exist_ok=True)
    cmd = [
        'colmap', 'model_converter',
        '--input_path', str(new_sparse_model_bin),
        '--output_path', str(new_sparse_model_txt),
        '--output_type', 'TXT'
    ]
    ret = subprocess.call(cmd)
    if ret != 0:
        logging.warning('Problem with model converter, exiting.')
        exit(ret)

    cmd = [
        'colmap', 'poisson_mesher',
        '--input_path',  str(dense_model / 'fused.ply'),
        '--output_path',  str(dense_model / 'meshed-poisson.ply')
    ]
    ret = subprocess.call(cmd)
    if ret != 0:
        logging.warning('Problem with poisson mesher, exiting.')
        exit(ret)

    cmd = [
        'colmap', 'delaunay_mesher',
        '--input_path',  str(dense_model),
        '--output_path',  str(dense_model / 'meshed-delaunay.ply'),
        '--input_type', 'dense'
    ]
    ret = subprocess.call(cmd)
    if ret != 0:
        logging.warning('Problem with delaunay mesher, exiting.')
        exit(ret)

    return dense_model


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--image_path', type=Path)
    parser.add_argument('--sparse_model', type=Path)
    args = parser.parse_args()
    main(**args.__dict__)
