#!/usr/bin/env python
import sys
sys.path.append('dataset_toolkit')
import os


from toc_dataset_toolkit import TOC_Dataset
from toc_dataset_toolkit.utils.logging import set_log_level
from renderer import resize_camera_parameters

from hloc.utils.database import COLMAPDatabase
from hloc.utils.read_write_model import CAMERA_MODEL_NAMES

import argparse
from pathlib import Path
import subprocess

import numpy as np
from pyquaternion import Quaternion


DENSE_DIRECTORY = 'dense_orb_model'
SPARSE_DIRECTORY = 'sparse_txt'

def main(dataset_root, scene_name, resize_factor, num_images):
    global DENSE_DIRECTORY
    set_log_level('WARNING')
    dataset = TOC_Dataset(root=dataset_root)
    scene_name_list = dataset.load_scene_name_list()
    scene_id = scene_name_list.index(scene_name)
    scene_dir = dataset_root / 'scenes' / scene_name
    keyframes_dir = scene_dir / DENSE_DIRECTORY / 'keyframes'

    dense_model_dir = scene_dir / DENSE_DIRECTORY
    sparse_model_dir = dense_model_dir / SPARSE_DIRECTORY

    triangulated_dir = dense_model_dir / 'triangulated'
    triangulated_dir.mkdir(exist_ok=True, parents=True)

    dense_ws = dense_model_dir / 'dense_ws'
    dense_ws.mkdir(exist_ok=True, parents=True)

    database_path = scene_dir / DENSE_DIRECTORY / 'database.db'

    database_path, imported_images  = create_colmap_database(
        dataset, scene_name, resize_factor, num_images
    )
    create_image_file(dataset, scene_name, imported_images)
    create_camera_file(dataset, scene_name, resize_factor)

    run_cmd([
        'colmap', 'feature_extractor',
        '--database_path', str(database_path),
        '--image_path', str(keyframes_dir),
    ])

    run_cmd([
        'colmap', 'exhaustive_matcher',
        '--database_path', str(database_path),
    ])

    run_cmd([
        'colmap', 'point_triangulator',
        '--database_path', str(database_path),
        '--image_path', str(keyframes_dir),
        '--input_path', str(sparse_model_dir),
        '--output_path', str(triangulated_dir),
    ])

    run_cmd([
        'colmap', 'image_undistorter',
        '--num_patch_match_src_images', str(5),
        '--image_path', str(keyframes_dir),
        '--input_path', str(triangulated_dir),
        '--output_path', str(dense_ws),
    ])

    run_cmd([
        'colmap', 'patch_match_stereo',
        '--workspace_path', str(dense_ws),
        '--PatchMatchStereo.cache_size', str(64),
        '--PatchMatchStereo.filter', str(1),
        '--PatchMatchStereo.depth_max', str(7),
        '--PatchMatchStereo.depth_min', str(0),
    ])

    run_cmd([
        'colmap', 'stereo_fusion',
        '--workspace_path', str(dense_ws),
        '--StereoFusion.min_num_pixels', str(50),
        '--output_path', str(dense_ws / 'fused.ply')
    ])

    run_cmd([
        'colmap', 'poisson_mesher',
        '--input_path',  str(dense_ws / 'fused.ply'),
        '--output_path',  str(dense_ws / 'meshed-poisson.ply')
    ])

    run_cmd([
        'colmap', 'delaunay_mesher',
        '--input_path',  str(dense_ws),
        '--output_path',  str(dense_ws / 'meshed-delaunay.ply'),
        '--input_type', 'dense'
    ])


def run_cmd(cmd):
    ret = subprocess.call(cmd)
    if ret != 0:
        print('Problem with ', ' '.join(cmd))
        exit(ret)


def create_colmap_database(dataset, scene_name, resize_factor, num_images):
    global DENSE_DIRECTORY

    scene_dir = dataset.root / 'scenes' / scene_name

    model_dir = scene_dir / DENSE_DIRECTORY
    model_dir.mkdir(exist_ok=True, parents=True)

    database_path = model_dir / 'database.db'
    if database_path.exists():
        # missing_ok only available from 3.8
        database_path.unlink() # missing_ok=True)

    database = create_empty_db(database_path)
    import_camera(dataset, scene_name, resize_factor, database)
    imported_images = import_images(dataset, scene_name, database, num_images)
    database.commit()
    database.close()

    return database_path, imported_images


def import_images(dataset, scene_name, database, num_images):
    global DENSE_DIRECTORY

    scene_dir = dataset.root / 'scenes' / scene_name
    img_dir = scene_dir / 'rgb_undistort'

    keyframes_dir = scene_dir / DENSE_DIRECTORY / 'keyframes'
    keyframes_dir.mkdir(exist_ok=True, parents=True)

    orb_poses_path = scene_dir / 'KeyFrameTrajectory.txt'
    orb_poses = np.loadtxt(orb_poses_path)

    total_images = len(orb_poses)
    if total_images <= num_images:
        step_size = 1
    else:
        step_size = int(np.floor(float(total_images) / num_images))
    print(f'total_images: {total_images}, num_images: {num_images}')
    print(f'step_size: {step_size}, num_images: {len(orb_poses[::step_size])}')

    selected_imgs = []
    for img_idx, orb_pose in enumerate(orb_poses[::step_size], 1):
        qw, qx, qy, qz, tx, ty, tz = orb_pose_to_colmap(orb_pose)
        img_name = f'color_{int(orb_pose[0])}.png'
        img_path = img_dir / img_name
        database.add_image(
            img_name, 1,
            np.asarray([qw, qx, qy, qz]),
            np.asarray([tx, ty, tz]), img_idx
        )
        symlink_force(img_path, keyframes_dir / img_name)
        selected_imgs.append({
            'path': keyframes_dir / img_name,
            'pose': (qw, qx, qy, qz, tx, ty, tz),
            'idx': img_idx
        })

    return selected_imgs


def symlink_force(target, link_name):
    try:
        os.symlink(target, link_name)
    except FileExistsError as e:
        os.remove(link_name)
        os.symlink(target, link_name)


def import_camera(dataset, scene_name, resize_factor, database):
    scene_id = dataset.load_scene_name_list().index(scene_name)
    camera_param = dataset.load_camera_param(scene_id)
    camera_param = resize_camera_parameters(camera_param, resize_factor)
    print(
        f'width: {camera_param.intrinsic.width}, '
        f'height: {camera_param.intrinsic.height}, '
        f'f: {camera_param.intrinsic.get_focal_length()}, '
        f'c: {camera_param.intrinsic.get_principal_point()}, '
    )
    fx, fy = camera_param.intrinsic.get_focal_length()
    cx, cy = camera_param.intrinsic.get_principal_point()
    width, height = camera_param.intrinsic.width, camera_param.intrinsic.height

    database.add_camera(
        CAMERA_MODEL_NAMES['PINHOLE'].model_id,
        camera_param.intrinsic.width, camera_param.intrinsic.height,
        [fx, fy, cx, cy], prior_focal_length=True
    )


def create_empty_db(database_path):
    db = COLMAPDatabase.connect(database_path)
    db.create_tables()
    db.commit()
    return db


def orb_pose_to_colmap(pose):
    x, y, z, qx, qy, qz, qw = pose[1:].tolist()
    T_world_cam = Quaternion(qw, qx, qy, qz).transformation_matrix
    T_world_cam[:3, -1] = [x, y, z]
    T_cam_world = np.linalg.inv(T_world_cam)
    quat = Quaternion(matrix=T_cam_world[:3, :3])
    x, y, z = T_cam_world[:3, -1].tolist()
    return quat[0], quat[1], quat[2], quat[3], x, y, z


def create_image_file(dataset, scene_name, images):
    global DENSE_DIRECTORY
    global SPARSE_DIRECTORY

    poses = [
        [img['idx'], *(img['pose']), 1, str(img['path'].name)]
        for img in images
    ]

    scene_dir = dataset.root / 'scenes' / scene_name
    sparse_model = scene_dir / DENSE_DIRECTORY / SPARSE_DIRECTORY
    sparse_model.mkdir(exist_ok=True, parents=True)

    lines = [
        '# Image list with two lines of data per image:',
        '#   IMAGE_ID, QW, QX, QY, QZ, TX, TY, TZ, CAMERA_ID, NAME',
        '#   POINTS2D[] as (X, Y, POINT3D_ID)',
        f'# Number of images: {len(poses)}, mean observations per image: 2'
    ]
    for pose in poses:
        lines.append(' '.join([str(i) for i in pose]))
        lines.append('')

    image_txt_file = sparse_model / 'images.txt'
    image_txt_file.write_text('\n'.join(lines)+'\n')

    points_txt_file = sparse_model / 'points3D.txt'
    points_txt_file.write_text('\n'.join([
        '# 3D point list with one line of data per point:',
        '#   POINT3D_ID, X, Y, Z, R, G, B, ERROR, TRACK[] as (IMAGE_ID, POINT2D_IDX)',
        '# Number of points: 0, mean track length: 1'
    ]))


def create_camera_file(dataset, scene_name, resize_factor):
    global DENSE_DIRECTORY
    global SPARSE_DIRECTORY

    scene_id = dataset.load_scene_name_list().index(scene_name)

    camera_param = dataset.load_camera_param(scene_id)
    camera_param = resize_camera_parameters(camera_param, resize_factor)
    print(
        f'width: {camera_param.intrinsic.width}, '
        f'height: {camera_param.intrinsic.height}, '
        f'f: {camera_param.intrinsic.get_focal_length()}, '
        f'c: {camera_param.intrinsic.get_principal_point()}, '
    )
    fx, fy = camera_param.intrinsic.get_focal_length()
    cx, cy = camera_param.intrinsic.get_principal_point()
    width, height = camera_param.intrinsic.width, camera_param.intrinsic.height

    scene_dir = dataset.root / 'scenes' / scene_name
    sparse_model = scene_dir / DENSE_DIRECTORY / SPARSE_DIRECTORY
    sparse_model.mkdir(exist_ok=True, parents=True)

    camera_txt_file = sparse_model / 'cameras.txt'
    camera_txt_file.write_text('\n'.join([
        '# Camera list with one line of data per camera:',
        '#   CAMERA_ID, MODEL, WIDTH, HEIGHT, PARAMS[]',
        '# Number of cameras: 1',
        f'1 PINHOLE {width} {height} {fx} {fy} {cx} {cy}'
    ]))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--dataset-root', type=Path,
        default='/root/3d_dataset'
    )
    parser.add_argument(
        '--scene-name', type=str,
        default='scene2'
    )
    parser.add_argument('--resize-factor', type=int, default=1)
    parser.add_argument('--num-images', type=int, default=150)
    args = parser.parse_args()
    main(**args.__dict__)
