#!/usr/bin/env python

import argparse
from pathlib import Path
import numpy as np
import yaml


def main(dataset_root, scene_name, resize_factor):
    path = dataset_root / 'scenes' / scene_name
    undistorted_path = path / 'rgb_undistort'
    intrinsics_path = path / 'color_camK.npy'
    new_params_path = path / 'orb_slam_params.yaml'
    new_intrinsics_path = path / 'intrinsics.txt'

    K = np.load(intrinsics_path).reshape((3, 3))
    fx = K[0, 0] / resize_factor
    fy = K[1, 1] / resize_factor
    cx = K[0, 2] / resize_factor
    cy = K[1, 2] / resize_factor
    print(fx, fy, cx, cy)

    width = int(1280 / resize_factor)
    height = int(720 / resize_factor)

    params = {
        'Camera.fx': float(fx),
        'Camera.fy': float(fy),
        'Camera.cx': float(cx),
        'Camera.cy': float(cy),

        'Camera.k1': 0.0,
        'Camera.k2': 0.0,
        'Camera.p1': 0.0,
        'Camera.p2': 0.0,
        'Camera.k3': 0.0,

        'Camera.width': width,
        'Camera.height': height,

        # Camera frames per second
        'Camera.fps': 30.0,

        # IR projector baseline times fx (aprox.)
        'Camera.bf': 40.0,

        # Color order of the images (0: BGR, 1: RGB. It is ignored if images are grayscale)
        'Camera.RGB': 1,

        # Close/Far threshold. Baseline times.
        'ThDepth': 40.0,

        # Deptmap values factor
        'DepthMapFactor': 1000.0,

        #--------------------------------------------------------------------------------------------
        # ORB Parameters
        #--------------------------------------------------------------------------------------------

        # ORB Extractor: Number of features per image
        'ORBextractor.nFeatures': 1000,

        # ORB Extractor: Scale factor between levels in the scale pyramid
        'ORBextractor.scaleFactor': 1.2,

        # ORB Extractor: Number of levels in the scale pyramid
        'ORBextractor.nLevels': 8,

        # ORB Extractor: Fast threshold
        # Image is divided in a grid. At each cell FAST are extracted imposing a minimum response.
        # Firstly we impose iniThFAST. If no corners are detected we impose a lower value minThFAST
        # You can lower these values if your images have low contrast
        'ORBextractor.iniThFAST': 10,
        'ORBextractor.minThFAST': 5,

        #--------------------------------------------------------------------------------------------
        # Viewer Parameters
        #--------------------------------------------------------------------------------------------
        'Viewer.KeyFrameSize': 0.05,
        'Viewer.KeyFrameLineWidth': 1,
        'Viewer.GraphLineWidth': 0.9,
        'Viewer.PointSize':2,
        'Viewer.CameraSize': 0.08,
        'Viewer.CameraLineWidth': 3,
        'Viewer.ViewpointX': 0,
        'Viewer.ViewpointY': -0.7,
        'Viewer.ViewpointZ': -1.8,
        'Viewer.ViewpointF': 500,
    }

    with open(str(new_params_path), 'w') as f:
        yaml.dump(params, f, default_flow_style=False)

    with open(str(new_params_path), 'r') as f:
        lines = f.readlines()
    lines = ['%YAML:1.0\n\n'] + lines

    with open(str(new_params_path), 'w') as f:
        f.write(''.join(lines))

    np.savetxt(str(new_intrinsics_path), np.asarray([
        fx, 0, cx,
        0, fy, cy,
        0, 0, 1
    ]).reshape((3, 3)))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '--dataset-root', type=Path,
        default='/root/3d_dataset'
    )
    parser.add_argument(
        '--scene-name', type=str,
        default='scene2'
    )
    parser.add_argument('--resize-factor', type=int, default=1)
    args = parser.parse_args()
    main(**args.__dict__)
