#!/usr/bin/env python

import sys
sys.path.append('dataset_toolkit')


from toc_dataset_toolkit import TOC_Dataset
from toc_dataset_toolkit.utils.logging import set_log_level

from align_utils import pose_align
from constants import get_workspace_bounding_box

import argparse
from pathlib import Path
import itertools
import time
import subprocess
import multiprocessing
import logging
import shutil
import re
import copy
import tempfile
import tqdm

import numpy as np
from scipy.spatial.transform import Rotation
import open3d as o3d
import pyrender
import trimesh
import matplotlib.pyplot as plt
import cv2


def main(dataset_root, scene_name):
    set_log_level('WARNING')
    dataset = TOC_Dataset(root=dataset_root)
    scene_name_list = dataset.load_scene_name_list()
    scene_id = scene_name_list.index(scene_name)
    print('scene', scene_id, scene_name)

    est_cam_poses = np.load(
        str(Path(dataset.root) / 'scenes' / scene_name / 'aligned_camera_poses.npy'),
        allow_pickle=True
    )[()]
    est_obj_poses = np.load(
        str(Path(dataset.root) / 'scenes' / scene_name / 'aligned_object_poses.npy'),
        allow_pickle=True
    )[()]

    occluded_masks, unoccluded_masks = render_object_masks(
        dataset, scene_id, est_cam_poses, est_obj_poses
    )
    selected_masks, rejected_masks = filter_severe_occlusions(
        occluded_masks, unoccluded_masks
    )
    segment_images(dataset, scene_id, rejected_masks)


def filter_severe_occlusions(occluded_masks, unoccluded_masks):
    selected_masks = {image_id: {} for image_id in occluded_masks.keys()}
    rejected_masks = {image_id: {} for image_id in occluded_masks.keys()}

    for image_id in occluded_masks.keys():
        for obj_name in occluded_masks[image_id].keys():
            occluded_mask = occluded_masks[image_id][obj_name]
            unoccluded_mask = unoccluded_masks[image_id][obj_name]
            iou = find_iou(occluded_mask, unoccluded_mask)
            masks = {
                'occluded': occluded_mask, 'unoccluded': unoccluded_mask,
                'iou': iou
            }
            if iou > 0.5:
                selected_masks[image_id][obj_name] = masks
            else:
                rejected_masks[image_id][obj_name] = masks

    return selected_masks, rejected_masks


def find_iou(mask1, mask2):
    intersection = np.logical_and(mask1, mask2)
    union = np.logical_or(mask1, mask2)
    iou_score = np.sum(intersection) / np.sum(union)
    return iou_score


def segment_images(dataset, scene_id, object_masks):
    for image_id, obj_masks in object_masks.items():
        rgb = dataset.load_raw_image(
            scene_id=scene_id, image_id=image_id, order = 'RGB'
        )
        for obj_name, obj_mask in obj_masks.items():
            alpha = (obj_mask['occluded'] != 0).astype(np.uint8) * 255
            rgba_occluded = np.concatenate((rgb, alpha[..., None]), axis=-1)

            alpha = (obj_mask['unoccluded'] != 0).astype(np.uint8) * 255
            rgba_unoccluded = np.concatenate((rgb, alpha[..., None]), axis=-1)

            if not np.isinf(obj_mask['iou']) and obj_mask['iou'] > 0.1:
                print(image_id, obj_name, obj_mask['iou'])

                plt.subplot(1, 3, 1)
                plt.imshow(rgb)
                plt.subplot(1, 3, 2)
                plt.imshow(rgba_occluded)
                plt.subplot(1, 3, 3)
                plt.imshow(rgba_unoccluded)
                plt.show()


def render_geometry(vis, geometry_list, camera_param):
    for g in geometry_list:
        vis.add_geometry(g)

    ctr = vis.get_view_control()
    camera_param.extrinsic = np.eye(4)
    ctr.convert_from_pinhole_camera_parameters(
        camera_param, allow_arbitrary=True
    )
    vis.poll_events()
    rendered_image = vis.capture_screen_float_buffer()
    rendered_image = np.asarray(rendered_image)

    for g in geometry_list:
        vis.remove_geometry(g)

    return rendered_image


def create_vis(camera_param):
    vis = o3d.visualization.Visualizer()
    vis.create_window(
        width=camera_param.intrinsic.width,
        height=camera_param.intrinsic.height,
        visible=True
    )
    return vis


def render_object_masks(dataset, scene_id, camera_poses, object_poses):
    occluded_masks = {}
    unoccluded_masks = {}

    object_meshes_world = {
        scene_object_name: \
            o3d.io.read_triangle_mesh(str(
                dataset.root / 'models' / scene_object_name / 'visual.ply'
            ))
        for scene_object_name in object_poses.keys()
    }

    camera_param = dataset.load_camera_param(scene_id)
    vis = create_vis(camera_param)

    for image_id, T_world_cam in tqdm.tqdm(camera_poses.items()):
        geometry_list = []
        object_masks = {}

        for scene_object_name, T_world_object in object_poses.items():
            if T_world_object is None:
                continue

            o3d_mesh_world = copy.deepcopy(
                object_meshes_world[scene_object_name]
            ).transform(T_world_object)
            o3d_mesh_cam = o3d_mesh_world.transform(np.linalg.inv(T_world_cam))
            if o3d_mesh_cam.has_textures():
                o3d_mesh_cam.texture.clear()

            obj_id = dataset.object_id_dict[scene_object_name]
            obj_mask_id = obj_id / 255.0
            o3d_mesh_cam = o3d_mesh_cam.paint_uniform_color(
                [obj_mask_id, obj_mask_id, obj_mask_id]
            )

            unoccluded_mask = render_geometry(vis, [o3d_mesh_cam], camera_param)
            unoccluded_mask = (unoccluded_mask[..., 0] == obj_mask_id)
            object_masks[scene_object_name] = unoccluded_mask

            geometry_list.append(o3d_mesh_cam)

        unoccluded_masks[image_id] = object_masks

        occluded_mask = render_geometry(vis, geometry_list, camera_param)

        object_masks = {}
        for scene_object_name in object_poses.keys():
            obj_id = dataset.object_id_dict[scene_object_name]
            obj_mask_id = obj_id / 255.0
            object_mask = (occluded_mask[..., 0] == (obj_mask_id))
            object_masks[scene_object_name] = object_mask

        occluded_masks[image_id] = object_masks

    vis.destroy_window()
    vis.close()

    return occluded_masks, unoccluded_masks


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--dataset-root', type=Path,
        default='/root/3d_dataset'
    )
    parser.add_argument(
        '--scene-name', type=str,
        default='20210512194607'
    )
    args = parser.parse_args()
    main(**args.__dict__)
