from trans3d import get_mat, get_pose,pose_4x4_to_pos_quat,pos_quat_to_pose_4x4
import open3d as o3d
import numpy as np
import cv2
import os
import copy
from xmlhandler import xmlReader,xmlWriter

FILTER_NUM_NEIGHBOR = 30
FILTER_STD_RATIO = 2.0
FILTER_RADIUS_M = 0.01

DOWNSAMPLE_VOXEL_SIZE_M = 0.002

ICP_MAX_DISTANCE_COARSE = DOWNSAMPLE_VOXEL_SIZE_M * 10
ICP_MAX_DISTANCE_FINE = DOWNSAMPLE_VOXEL_SIZE_M * 1

NORMAL_RADIUS = 0.01
NORMAL_NUM_NEIGHBOR = 30

PLANE_Z_MIN = -0.005
PLANE_Z_MAX = 0.002
PLANE_Y_MIN = -0.31
PLANE_Y_MAX = 0
PLANE_X_MIN = 0
PLANE_X_MAX = 0.29


MARKER_LENGTH = 25

MAX_DEPTH_M = 0.7
MIN_DEPTH_M = 0.1

WIDTH = 1280
HEIGHT = 720

DOWNSAMPLE_VOXEL_SIZE_M = 0.005



def convert_camera_poses(cameraposes):
    # because there are two versions of camera poses npy files.
    # one is the (:,7) shape as [pos(3),quat(4)]
    # the other is the (:,4,4) shape that specifies the 4x4 pose matrix
    if cameraposes.shape[1] == 4:
        return cameraposes
    # the other shape
    elif cameraposes.shape[1] == 7:
        transformed_camera_poses = np.empty(shape=(cameraposes.shape[0],4,4))
        for i in range(len(cameraposes)):
            translation = cameraposes[i][0:3]
            quat = cameraposes[i][3:7]
            transformed_camera_poses[i] = pos_quat_to_pose_4x4(translation,quat)
        return transformed_camera_poses
    else:
        # other shape is not accepted
        raise ValueError('Shape for camerapose should be either (:,7) or (:,4,4)')

def rgbd_with_mask_to_pointcloud(color_image_path, depth_image_path, mask_path, width, height, camera_matrix, object_id):
    '''
    Input:
    color_image_path: the path of the rgb image
    depth_image_path: the path of the depth image
    mask_path: the path of the npy mask file
    width: the width of the image
    height: the height of the image
    camera_matrix: the 3x3 numpy camera intrinsic matrix
    object_id: the object id of the object

    Output:
    Open3d geometry class which is the selected part of the point cloud
    '''
    color = o3d.io.read_image(color_image_path)
    depth = cv2.imread(depth_image_path, cv2.IMREAD_UNCHANGED).astype(np.float32)
    mask = np.load(mask_path)
    depth /= 1000.0
    depth[depth < MIN_DEPTH_M] = 0
    depth[depth > MAX_DEPTH_M] = 0
    # print(depth.dtype)
    mask_01 = (mask == object_id).astype(int)
    # print(mask_01)

    depth = (depth * mask_01).astype(np.float32)
    
    rgbd_image = o3d.geometry.RGBDImage()
    rgbd_image.color = color
    rgbd_image.depth = o3d.geometry.Image(depth)

    intrinsic = o3d.camera.PinholeCameraIntrinsic()
    intrinsic.set_intrinsics(
        width, height, camera_matrix[0, 0], camera_matrix[1, 1], camera_matrix[0, 2], camera_matrix[1, 2])
    pcd = o3d.geometry.PointCloud()
    pcd = o3d.geometry.PointCloud.create_from_rgbd_image(rgbd_image, intrinsic, np.eye(4))
    # pcd = o3d.geometry.create_point_cloud_from_rgbd_image(rgbd_image, intrinsic, np.eye(4))
    return pcd

def transform_pcd(pcd,transformation_matrix):
    '''
    Input:
    pcd: point cloud in the format of open3d point cloud class
    transformation_matrix: the 4x4 numpy array of the transformation matrix

    Output:
    the point cloud after transformation
    '''
    pcd = pcd.transform(transformation_matrix)
    return pcd

def refine_6d_pose(scene, model, x, y ,z, alpha, beta, gamma):
    '''
    Input:
    scene: the open3d geometry class instance
    model: the open3d geometry class instance
    x,y,z,alpha,beta,gamma: the pose of the model relative to the scene
    
    Output:
    the refined pose compose of x,y,z,alpha,beta,gamma
    '''
    print("log:icp")
    current_transformation = get_mat(x, y ,z, alpha, beta, gamma)

    radius_normal = DOWNSAMPLE_VOXEL_SIZE_M * 2

    model.estimate_normals(o3d.geometry.KDTreeSearchParamHybrid(radius=radius_normal, max_nn=50))
    scene.estimate_normals(o3d.geometry.KDTreeSearchParamHybrid(radius=radius_normal, max_nn=50))

    radius_feature = DOWNSAMPLE_VOXEL_SIZE_M * 5
    model_fpfh = o3d.registration.compute_fpfh_feature(model, o3d.geometry.KDTreeSearchParamHybrid(radius=radius_feature, max_nn=100))
    scene_fpfh = o3d.registration.compute_fpfh_feature(scene, o3d.geometry.KDTreeSearchParamHybrid(radius=radius_feature, max_nn=100))

    result_icp = o3d.registration.registration_icp(
            model, scene, DOWNSAMPLE_VOXEL_SIZE_M, current_transformation,
            o3d.registration.TransformationEstimationPointToPoint())
    # o3d.io.write_point_cloud('./scene_trans.ply', scene, write_ascii=True)
    # o3d.io.write_point_cloud('./model.ply', copy.deepcopy(model).transform(current_transformation), write_ascii=True)

    print("log:icp done")
    # matrix -> 6d
    x,y,z, alpha, beta, gamma = get_pose(result_icp.transformation)
    return x,y,z, alpha, beta, gamma

    
def load_transformed_scenes(cam,rgb_image_path,depth_image_path,mask_path,pose,object_id_list):
    '''
    Input:
    cam: camera type: kinect or realserse
    rgb_image_path: the path of the image file
    depth_image_path: the path of the depth image file
    mask_path: the mask of the npy mask file
    pose:the 4x4 numpy array pose of the camera
    object_id_list: the list of the object ids
    
    Output:
    A list of pcds of the objects that specified by the object_id_list
    '''
    if cam == 'realsense':
        camera_matrix = np.array([927.17,0.0,651.51506348,0.0,927.36688232,349.62133789,0.0,0.0,1.0]).reshape([3,3])
    elif cam == 'kinect':
        camera_matrix = np.array([[631.54864502,0.,638.43517329],[0.,631.20751953,366.49904066],[0.,0.,1.]])
    else:
        raise ValueError('Unknown camera type %s' % cam)
    pcds = []
    for object_id in object_id_list:
        pcd = rgbd_with_mask_to_pointcloud(rgb_image_path, depth_image_path, mask_path, 1280, 720, camera_matrix,object_id)
        ### downsample
        # pcd.voxel_down_sample(DOWNSAMPLE_VOXEL_SIZE_M)
        ### estimate normals
        pcd.estimate_normals(search_param = o3d.geometry.KDTreeSearchParamHybrid(radius = NORMAL_RADIUS,max_nn = NORMAL_NUM_NEIGHBOR))

        # pcd.transform(np.linalg.inv(pose))
        pcds.append(pcd)
    return pcds

object_file_name_list_file = open('object_file_name_list.txt','r')
lines = object_file_name_list_file.readlines()
objectfilenamelist = []
for line in lines:
    if not (line == '\n'):
        objectfilenamelist.append(line.replace('\n','').replace('\r',''))
print('log:loaded object file name list:',end='')
print(objectfilenamelist)

object_id_list_file = open('scene_33/object_id_list.txt','r')
lines = object_id_list_file.readlines()
object_id_list=[]
for line in lines:
    if not (line == '\n'):
        object_id_list.append(int(line))
print('log:loaded object id list:',end='')
print(object_id_list)

image_path = 'scene_33/data/rgb/0000.png'
depth_path = 'scene_33/data/depth/0000.png'
mask_path = 'scene_33/data/masks/0000.npy'
poses = np.load('scene_33/data/camera_poses.npy')
poses = convert_camera_poses(poses)

models_o3d = []
for object_id in object_id_list:
    models_o3d.append(o3d.io.read_point_cloud(os.path.join('models',objectfilenamelist[object_id])))

main_xml_reader = xmlReader('scene_33/data/annotations/0000.xml')
pose_vector_list = main_xml_reader.getposevectorlist()

transformed_scenes = load_transformed_scenes('realsense',image_path,depth_path,mask_path,poses[0],object_id_list)

if not len(pose_vector_list) == len(transformed_scenes):
    raise ValueError('The length of pose vector list does not equal to the length of the transformed scenes')

num_objects = len(pose_vector_list)

for i in range(num_objects):
    object_id,x,y,z,alpha,beta,gamma = pose_vector_list[i]
    x0,y0,z0,alpha0,beta0,gamma0 = x,y,z,alpha,beta,gamma


    print('objectid:%d\nBefore,x=%.6f,y=%.6f,z=%.6f,alpha=%.6f,beta=%.6f,gamma=%.6f' % (object_id,x,y,z,alpha,beta,gamma))
    x,y,z,alpha,beta,gamma = refine_6d_pose(transformed_scenes[i], models_o3d[i], x, y ,z, alpha, beta, gamma)
    print('After,x=%.6f,y=%.6f,z=%.6f,alpha=%.6f,beta=%.6f,gamma=%.6f' % (x,y,z,alpha,beta,gamma))
    

    vis = o3d.visualization.Visualizer()
    vis.create_window(width = 1280,height = 720)
    vis.add_geometry(transformed_scenes[i])
    model = copy.deepcopy(models_o3d[i])
    model.transform(get_mat(x0,y0,z0,alpha0,beta0,gamma0))
    vis.add_geometry(model)
    vis.run()
    vis.destroy_window()

    vis = o3d.visualization.Visualizer()
    vis.create_window(width = 1280,height = 720)
    vis.add_geometry(transformed_scenes[i])
    model = copy.deepcopy(models_o3d[i])
    model.transform(get_mat(x,y,z,alpha,beta,gamma))
    vis.add_geometry(model)
    vis.run()
    vis.destroy_window()