#!/usr/bin/env python
import sys
sys.path.append('dataset_toolkit')

from toc_dataset_toolkit import TOC_Dataset
from toc_dataset_toolkit.utils.logging import set_log_level

import torch
import torch.nn.functional as F
from torchviz import make_dot

from pytorch3d.io import load_objs_as_meshes
from pytorch3d.renderer.cameras import PerspectiveCameras
from pytorch3d.structures import join_meshes_as_batch

from pytorch3d.renderer import (
    look_at_view_transform,
    PerspectiveCameras,
    FoVPerspectiveCameras,
    SoftPhongShader,
    SoftSilhouetteShader,
    RasterizationSettings,
    BlendParams,
    MeshRenderer,
    MeshRasterizer,
    PointLights,
    AmbientLights
)
import pytorch3d.transforms

import argparse
from pathlib import Path
import numpy as np
import time
import tqdm
import itertools
import random

import matplotlib.pyplot as plt
import open3d as o3d
import pyrender
import trimesh
import cv2


# Pytorch3D renderer doesn't work with width != height
# plus reduce image size for memory reasons
NEW_WIDTH = 1000
NEW_HEIGHT = 1000
NEW_FOCAL_LENGTH_X = NEW_FOCAL_LENGTH_Y = 850
NEW_PRINCIPAL_POINT_X = NEW_PRINCIPAL_POINT_Y = 500

NUM_EPOCHS = 500
BATCH_SIZE = 4
# number of iterations without loss drop to declare convergence
MAX_NO_LOSS_DECREASE_ITERS = 50


def main(dataset_root, scene_name):
    set_log_level('WARNING')

    dataset = TOC_Dataset(root=dataset_root)
    scene_name_list = dataset.load_scene_name_list()
    scene_id = scene_name_list.index(scene_name)
    # scene_name = scene_name_list[scene_id]

    print('scene', scene_id, scene_name)

    est_cam_poses = np.load(
        str(Path(dataset.root) / 'scenes' / scene_name / 'aligned_camera_poses.npy'),
        allow_pickle=True
    )[()]
    est_obj_poses = np.load(
        str(Path(dataset.root) / 'scenes' / scene_name / 'aligned_object_poses.npy'),
        allow_pickle=True
    )[()]

    if True:
        # object_name = 'suger_1'
        object_name = 'suger_2'
        # object_name = 'potato_chip_3'
        object_pose = est_obj_poses[object_name]
    # for object_name, object_pose in est_obj_poses.items()
        refine_object_pose(
            dataset, scene_id, scene_name,
            object_name, object_pose, est_cam_poses
        )


def refine_object_pose(
    dataset, scene_id, scene_name, object_name, object_pose, camera_poses
):
    device = 'cuda'

    camera_param = dataset.load_camera_param(scene_id)
    old_fx, old_fy = camera_param.intrinsic.get_focal_length()
    old_cx, old_cy = camera_param.intrinsic.get_principal_point()
    old_K = K_matrix(old_fx, old_fy, old_cx, old_cy)

    new_K = K_matrix(
        NEW_FOCAL_LENGTH_X, NEW_FOCAL_LENGTH_Y,
        NEW_PRINCIPAL_POINT_X, NEW_PRINCIPAL_POINT_Y,
    )

    rectification_masks = get_rectification_masks(
        old_K, new_K,  (NEW_WIDTH, NEW_HEIGHT)
    )

    renderer = get_pytorch3d_renderer(device, NEW_WIDTH, NEW_HEIGHT)

    # se3_log_increment = torch.full(
    #     (1, 6), 0.0, device=device, requires_grad=False
    # )
    se3_log_increment = torch.randn(
        1, 6, dtype=torch.float32, device=device
    )
    se3_log_increment[0, :3] *= 0.02 # 1e-5 #
    se3_log_increment[0, 3:6] *= 0.5 # 1e-5 #
    print('se3 noise: ', se3_log_increment.tolist())

    T_world_obj = torch.from_numpy(object_pose).float().unsqueeze(0).to(device)
    gt_T_world_obj = T_world_obj.clone()

    mesh_path = dataset.root / 'models' / object_name / \
                    'textured.obj'
    mesh = load_objs_as_meshes([mesh_path], device=device)
    # same mesh for all camera positions
    mesh = join_meshes_as_batch([mesh for _ in range(BATCH_SIZE)])

    gt_object_masks = {}
    # for cam_idx in [5]:
    gt_masks_path = Path(dataset.root / 'scenes' / scene_name / 'gt_masks')
    gt_masks_path.mkdir(exist_ok=True)

    # hand picked cam poses for suger_2 object
    # TODO: select cam_indices automatically based on occlusion
    cam_indices = list(range(0, 12)) \
                + [21, 24, 34, 44, 46, 47, 48, 50, 51] \
                + list(range(53, 57)) \
                + list(range(59, 66))
    # cam_indices = [10]

    def chunk(lst, n):
        """divides list into even sized chunks"""
        return [lst[i:i + n] for i in range(0, len(lst), n)]

    random.shuffle(cam_indices)
    cam_indices_batched = chunk(cam_indices, BATCH_SIZE)

    print('loading gt masks')
    for cam_idx in tqdm.tqdm(cam_indices):
        mask_file = gt_masks_path / (object_name + '_' + str(cam_idx) + '.npy')

        # gt_object_masks[cam_idx] = get_object_mask(
        #     mesh_path, object_pose, camera_poses[cam_idx]
        # )
        # np.save(mask_file, gt_object_masks[cam_idx])

        gt_object_masks[cam_idx] = np.load(mask_file)

    # se3_log_increment.requires_grad = True
    # optimizer = torch.optim.SGD(
    #     [se3_log_increment], lr=1e-3, momentum=0.0
    # )

    best_loss = np.inf
    no_loss_decrease_iters = 0
    losses = []

    for epoch_idx, minibatch_idx in tqdm.tqdm(list(itertools.product(
        range(NUM_EPOCHS), range(len(cam_indices_batched))
    ))):
        if no_loss_decrease_iters >= MAX_NO_LOSS_DECREASE_ITERS:
            print('loss failed to decrease for %d iters, terminating'
                        % no_loss_decrease_iters)
            break

        cam_batch_indices = cam_indices_batched[minibatch_idx]
        # if object mask small cuz, don't bother
        for cam_idx in cam_batch_indices:
            if np.sum(gt_object_masks[cam_idx].astype(float)) < 10000:
                continue

        if torch.any(torch.isnan(se3_log_increment)):
            print('nan se3_log_increment')
            se3_log_increment = torch.zeros_like(se3_log_increment.detach())

        # log maps should not get too large.
        # So update the pose and set increment to zero
        T_increment = se3_exp_map(se3_log_increment).detach()
        T_world_obj = (T_world_obj @ T_increment).detach()

        trans_increment = torch.zeros(
            1, 3, dtype=torch.float32, device=device, requires_grad=True
        )
        so3_log_increment = torch.zeros(
            1, 3, dtype=torch.float32, device=device, requires_grad=True
        )
        se3_log_increment = torch.cat(
            (trans_increment, so3_log_increment), dim=-1
        )

        # find if there is a better way instead of re-creating optimizer every time
        optimizer = torch.optim.SGD(
            [
                {'params': trans_increment, 'lr': 1e-2},
                {'params': so3_log_increment, 'lr': 5e-1}
            ],
            lr=1e-2, momentum=0.0
        )
        optimizer.zero_grad()

        # need this to link the loss gradient to se3_log_increment
        T_increment = se3_exp_map(se3_log_increment)
        T_world_obj = T_world_obj @ T_increment

        original_images, original_depths = get_original_images(
            dataset, scene_name, cam_batch_indices, rectification_masks
        )
        original_images = original_images.to(device)
        original_depths = original_depths.to(device)

        rendered_images, rendered_depths = render_object(
            dataset, scene_name, renderer, object_name, T_world_obj,
            cam_batch_indices, camera_poses, mesh, original_images
        )

        rendered_masks = rendered_images[..., -1] \
                            .unsqueeze(-1) \
                            .expand(-1, -1, -1, 3)

        # if object mask too small, don't bother
        if torch.sum((rendered_masks > 1e-7).float()).item() < 10000:
            continue

        gt_masks = []
        for cam_idx in cam_batch_indices:
            gt_masks.append(torch.from_numpy(gt_object_masks[cam_idx]) \
                                .unsqueeze(0).unsqueeze(-1) \
                                .expand(-1, -1, -1, 3) \
                                .float().to(device))
        gt_masks = torch.cat(gt_masks, dim=0)

        original_images_masked = original_images * gt_masks
        rendered_images_masked = rendered_images[..., :3] * rendered_masks

        original_depths_masked = original_depths * gt_masks[:, :, :, 0]
        rendered_depths_masked = rendered_depths * (rendered_depths > 0).float()
        loss = torch.mean(torch.abs(
            # (original_images_masked - rendered_images_masked) ** 2
            # original_depths_masked - rendered_depths_masked
            (gt_masks[..., 0] - rendered_masks[..., 0]) ** 2
        ))

        if torch.isnan(loss):
            print('nan loss')
            continue

        loss.backward()
        optimizer.step()

        se3_log_increment = torch.cat(
            (trans_increment, so3_log_increment), dim=-1
        )

        if torch.any(torch.isnan(se3_log_increment)):
            print('nan se3_log_increment')
            no_loss_decrease_iters += 1
            continue

        T_increment = se3_exp_map(se3_log_increment.detach())
        T_world_obj_new = T_world_obj @ T_increment
        err0 = get_pose_diff(gt_T_world_obj, T_world_obj)
        err1 = get_pose_diff(gt_T_world_obj, T_world_obj_new)

        losses.append(
            [loss.item(), *(get_se3_log_magnitude(err0))]
        )

        if loss.item() < best_loss:
            no_loss_decrease_iters = 0
            best_loss = loss.item()

            print(
                'epoch:', epoch_idx,
                'cam:', minibatch_idx,
                'prev err:', get_se3_log_magnitude(err0),
                'curr err:', get_se3_log_magnitude(err1),
                'loss: ', loss.item(),
                'inc:', get_se3_log_magnitude(se3_log_increment),
                'grad: ', [
                    torch.norm(trans_increment.grad[0]).item() \
                       if (trans_increment.grad is not None) else None,
                    torch.norm(so3_log_increment.grad[0]).item() \
                        if (so3_log_increment.grad is not None) else None
                ]
            )

            plt.figure(figsize=(10, 10))
            for cam_rel_idx, cam_idx in enumerate(cam_batch_indices):
                num_cols = 3
                plt_offset = cam_rel_idx * num_cols

                plt.subplot(BATCH_SIZE, num_cols, 1 + plt_offset)
                plt.imshow(
                    original_images[cam_rel_idx, ..., :3] \
                        .detach().cpu().numpy()
                )

                plt.subplot(BATCH_SIZE, num_cols, 2 + plt_offset)
                plt.imshow(
                    original_images_masked[cam_rel_idx, ..., :3] \
                        .detach().cpu().numpy()
                )
                # plt.imshow(original_depths_masked[cam_rel_idx].detach().cpu().numpy())
                # plt.imshow(
                #     gt_masks[cam_rel_idx, :, :, 0].detach().cpu().numpy()
                # )

                plt.subplot(BATCH_SIZE, num_cols, 3 + plt_offset)
                plt.imshow(
                    original_images_masked[cam_rel_idx, ..., :3] \
                        .detach().cpu().numpy()
                )
                plt.imshow(
                    rendered_masks[cam_rel_idx, :, :, 0].detach().cpu().numpy(),
                    alpha=0.5, cmap='gray', vmin=0, vmax=1.0
                )

                # plt.subplot(BATCH_SIZE, num_cols, 4 + plt_offset)
                # plt.imshow(
                #     rendered_images_masked[cam_rel_idx, ..., :3] \
                #         .detach().cpu().numpy()
                # )
                # # plt.imshow(rendered_depths[cam_rel_idx].detach().cpu().numpy())
                # plt.imshow(rendered_masks[cam_rel_idx, :, :, 0].detach().cpu().numpy())

            plt.savefig('%d_%d_%s.png' % (
                epoch_idx, minibatch_idx, '-'.join(map(str, cam_batch_indices))
            ))
            # plt.show()
            plt.close()
        else:
            no_loss_decrease_iters += 1

        # make_dot(
        #     loss, params={'se3_log_increment': se3_log_increment},
        #     show_attrs=True, show_saved=True
        # ).render("torchviz", format="png")
    np.savetxt('losses.npy.txt', losses)


def get_original_images(
    dataset, scene_name, cam_batch_indices, rectification_masks
):
    original_images = []
    original_depths = []

    for cam_idx in cam_batch_indices:
        original_image_path \
            = dataset.root / 'scenes' / scene_name \
                        / 'rgb_undistort' / ('%.4d.png' % cam_idx)
        original_image = cv2.imread(str(original_image_path))
        original_image = rectify_image(original_image, rectification_masks)
        original_image = torch.from_numpy(
            original_image[..., ::-1].copy()
        ).float().unsqueeze(0) / 255.
        original_images.append(original_image)

        original_depth_path \
            = dataset.root / 'scenes' / scene_name \
                        / 'depth_undistort' / ('%.4d.png' % cam_idx)
        original_depth = cv2.imread(str(original_depth_path), -1)
        original_depth = rectify_image(original_depth, rectification_masks)
        original_depth = torch.from_numpy(
            original_depth.astype(float)
        ).float().unsqueeze(0) / 1000.
        original_depths.append(original_depth)

    original_images = torch.cat(original_images, dim=0)
    original_depths = torch.cat(original_depths, dim=0)
    return original_images, original_depths


def get_colorized_depth(depth, mask):
    colorized_depth = cv2.applyColorMap(
        (255 * 0.5 * depth[0]) \
            .detach().cpu().numpy().astype(np.uint8),
        cv2.COLORMAP_JET
    )
    return colorized_depth * mask[0].detach().cpu().numpy().astype(np.uint8)


def se3_exp_map(se3):
    # transpose cuz pytorch3d's weird
    return pytorch3d.transforms.se3_exp_map(se3).permute(0, 2, 1)


def se3_log_map(SE3):
    # transpose cuz pytorch3d's weird
    return pytorch3d.transforms.se3_log_map(SE3.permute(0, 2, 1))


def get_se3_log_magnitude(se3):
    return [torch.norm(se3[0, :3]).item(), torch.norm(se3[0, 3:]).item()]


def get_pose_diff(pose1, pose2):
    pose_1_2 = torch.inverse(pose1) @ pose2
    return se3_log_map(pose_1_2)


def render_object(
    dataset, scene_name, renderer, object_name, T_world_obj,
    cam_batch_indices, camera_poses, mesh, original_images
):
    device = renderer.lights.device

    focal_length_tensor = torch.tensor(
        [NEW_FOCAL_LENGTH_X, NEW_FOCAL_LENGTH_Y], device=device
    ).unsqueeze(0)
    principal_point_tensor = torch.tensor(
        [NEW_PRINCIPAL_POINT_X, NEW_PRINCIPAL_POINT_Y], device=device
    ).unsqueeze(0)

    T_world_camera = []
    for cam_idx in cam_batch_indices:
        T_world_camera.append(torch.from_numpy(
            camera_poses[cam_idx]
        ).float().unsqueeze(0).to(device))
    T_world_camera = torch.cat(T_world_camera, dim=0)
    T_world_objs = T_world_obj.expand(len(cam_batch_indices), -1, -1)

    T_obj_camera = torch.inverse(T_world_objs) @ T_world_camera
    T_obj_camera_pytorch = change_pose_to_pytorch3d(T_obj_camera)
    T_camera_obj_pytorch3d = torch.inverse(T_obj_camera_pytorch)

    # rotation transposed because PyTorch3D's weird convention
    R = T_camera_obj_pytorch3d[:, :3, :3].permute(0, 2, 1)
    T = T_camera_obj_pytorch3d[:, :3, -1]

    camera = PerspectiveCameras(
        focal_length=focal_length_tensor,
        principal_point=principal_point_tensor,
        image_size=((NEW_WIDTH, NEW_HEIGHT), ), # (W, H)
        R=R, T=T, device=device
    )

    rendered_images, rendered_depths = renderer(
        mesh, cameras=camera, lights=renderer.lights
    )

    ## debug
    # T_obj_camera_np = T_obj_camera[0].detach().cpu().numpy()

    # pyrendered_image, _ = pyrender_image(
    #     mesh_path, T_obj_camera_np,
    #     NEW_FOCAL_LENGTH_X, NEW_FOCAL_LENGTH_Y,
    #     NEW_PRINCIPAL_POINT_X, NEW_PRINCIPAL_POINT_Y,
    #     NEW_WIDTH, NEW_HEIGHT
    # )

    # mesh_path = dataset.root / 'models' / object_name / \
    #                 'textured.obj'
    # o3d_mesh = o3d.io.read_triangle_mesh(
    #     str(mesh_path), enable_post_processing=True, print_progress=True
    # )
    # o3d.visualization.draw_geometries([
    #     o3d_mesh,
    #     o3d.geometry.TriangleMesh.create_coordinate_frame(0.1).transform(T_obj_camera_np)
    # ])
    # exit(0)

    return rendered_images, rendered_depths


class MeshRendererWithDepth(torch.nn.Module):
    def __init__(self, rasterizer, shader, lights):
        super().__init__()
        self.rasterizer = rasterizer
        self.shader = shader
        self.lights = lights

    def forward(self, meshes_world, **kwargs) -> torch.Tensor:
        fragments = self.rasterizer(meshes_world, **kwargs)
        images = self.shader(fragments, meshes_world, **kwargs)
        return images, fragments.zbuf[..., 0]


def get_pytorch3d_renderer(device, width, height):
    lights = AmbientLights(device=device, ambient_color=[[0.75, 0.75, 0.75]])
    blend_params = BlendParams(sigma=5e-6, gamma=1e-10)
    raster_settings = RasterizationSettings(
        image_size=(height, width),
        blur_radius=np.log(1. / 1e-4 - 1.) * blend_params.sigma,
        # low faces per pixel ok if not using photometric error
        faces_per_pixel=3,
        # faces_per_pixel=20,
        perspective_correct=True
    )

    renderer = MeshRendererWithDepth(
        rasterizer=MeshRasterizer(
            raster_settings=raster_settings
        ),
        shader=SoftSilhouetteShader(),
        # loss doesn't decrease using phong shader
        # shader=SoftPhongShader(
        #     device=device,
        #     lights=lights,
        #     blend_params=blend_params
        # ),
        lights=lights
    )

    return renderer


def pyrender_image(mesh_path, cam_pose, fx, fy, cx, cy, width, height):
    cam = pyrender.IntrinsicsCamera(fx=fx, fy=fy, cx=cx, cy=cy)

    trimesh_mesh = trimesh.load(mesh_path)
    scene = pyrender.Scene(
        ambient_light=np.array([0.35, 0.35, 0.35, 1.0])
    )
    pyrender_mesh = pyrender.Mesh.from_trimesh(trimesh_mesh)
    mesh_node = scene.add(pyrender_mesh)

    cam_pose = change_pose_to_opengl(cam_pose)
    cam_node = scene.add(cam, pose=cam_pose)
    r = pyrender.OffscreenRenderer(
        viewport_width=width,
        viewport_height=height
    )
    rendered_color, rendered_depth = r.render(scene)
    r.delete()
    return rendered_color, rendered_depth


def get_object_mask(mesh_path, object_pose, camera_pose):
    T_world_obj = object_pose
    T_world_camera = camera_pose
    T_obj_camera = np.linalg.inv(T_world_obj).dot(T_world_camera)

    _, rendered_depth = pyrender_image(
        mesh_path, T_obj_camera,
        NEW_FOCAL_LENGTH_X, NEW_FOCAL_LENGTH_Y,
        NEW_PRINCIPAL_POINT_X, NEW_PRINCIPAL_POINT_Y,
        NEW_WIDTH, NEW_HEIGHT
    )

    return rendered_depth > 1e-7

def get_rectification_masks(old_intrinsics, new_intrinsics, new_size):
    return cv2.initUndistortRectifyMap(
        old_intrinsics, newCameraMatrix=new_intrinsics,
        distCoeffs=None, size=new_size, R=None, m1type=cv2.CV_32FC1
    )


def rectify_image(image, rectification_masks):
    return cv2.remap(
        image, rectification_masks[0], rectification_masks[1], cv2.INTER_LINEAR
    )


def K_matrix(fx, fy, cx, cy):
    return np.asarray([
        [fx, 0, cx],
        [0, fy, cy],
        [0, 0, 1]
    ])


def change_pose_to_pytorch3d(pose):
    '''to PyTorch3D convention (West-Up-North) from East-Down-North'''
    rotation_correction = torch.tensor(
        [[[-1, 0, 0], [0, -1, 0], [0, 0, 1]]],
        device=pose.device, dtype=pose.dtype
    ).expand(pose.shape[0], -1, -1)

    pose = torch.clone(pose)
    pose[:, :3, :3] = pose[:, :3, :3] @ rotation_correction
    return pose


def change_pose_to_opengl(pose):
    '''to OpenGL convention (East-Up-South) from East-Down-North'''
    rotation_correction = np.asarray([
        [1, 0, 0], [0, -1, 0], [0, 0, -1]
    ])
    pose = pose.copy()
    pose[:3, :3] = pose[:3, :3].dot(rotation_correction)
    return pose


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--dataset-root', type=Path,
        default='/root/3d_dataset'
    )
    parser.add_argument(
        '--scene-name', type=str,
        default='20210512194607'
    )
    args = parser.parse_args()
    main(**args.__dict__)
