#!/usr/bin/env python
import sys
sys.path.append('dataset_toolkit')


from toc_dataset_toolkit import TOC_Dataset
from toc_dataset_toolkit.utils.logging import set_log_level
from renderer import resize_camera_parameters

from align_utils import pose_align
from constants import get_workspace_bounding_box

import argparse
from pathlib import Path
import itertools
import time
import subprocess
import multiprocessing
import logging
import shutil
import re
import copy
import tempfile
import tqdm

import numpy as np
from scipy.spatial.transform import Rotation
import open3d as o3d
import pyrender
import trimesh
import matplotlib.pyplot as plt
import cv2

# in degree
ROTATION_ERR_THRESH = 1.5

# in meters
TRANSLATION_ERR_THRESH = 0.01

# in meters
ROBUST_ALIGNMENT_MAX_ERR = 0.005

# number of points to sample from a mesh
SCENE_MESH_SAMPLE_POINTS = 500000
OBJECT_MESH_SAMPLE_POINTS = 100000

# ICP alignment threshold
ICP_ALIGNMENT_THRESHOLD = 0.05


def main(dataset_root, scene_name, resize_factor):
    set_log_level('WARNING')
    dataset = TOC_Dataset(root=dataset_root)
    scene_name_list = dataset.load_scene_name_list()
    scene_id = scene_name_list.index(scene_name)

    print('scene', scene_id, scene_name)

    colmap_output = dataset_root / 'scenes' / scene_name_list[scene_id] \
                        / 'colmap_output_all_images_2'
    pose_filename = colmap_output / 'images.txt'
    points_filename = colmap_output / 'points3D.txt'

    camera_poses, rendered_est_poses = parse_colmap_poses(pose_filename)

    rendered_gt_poses = {}
    for scene_object_name in rendered_est_poses.keys():
        rendered_gt_poses[scene_object_name] = get_gt_object_poses(
            dataset, scene_name_list, scene_id, scene_object_name
        )

    aligned_colmap_output, inliers = align_objects_sim_transforms(
        colmap_output, rendered_gt_poses
    )
    pose_filename = aligned_colmap_output / 'images.txt'
    points_filename = aligned_colmap_output / 'points3D.txt'
    camera_poses, rendered_est_poses = parse_colmap_poses(pose_filename)
    point_cloud = parse_colmap_points(points_filename)

    object_poses = find_object_poses(
        rendered_gt_poses, rendered_est_poses, inliers
    )

    out_scene_dir=Path(dataset.root)/ 'scenes'/ scene_name_list[scene_id]
    np.save(str(out_scene_dir/'camera_poses.npy'), camera_poses)
    np.save(str(out_scene_dir/'object_poses.npy'), object_poses)

    viz_objects(
        dataset, scene_id, object_poses, resize_factor,
        camera_poses, rendered_est_poses, point_cloud
    )
    exit(0)

    dense_colmap_output = dataset_root / 'scenes' / scene_name_list[scene_id] \
                        / 'dense_model2'

    dense_colmap_output = do_dense_reconstruction(
        dataset, scene_name_list[scene_id],
        aligned_colmap_output, # None, #
        dense_colmap_output
    )

    shutil.rmtree(aligned_colmap_output)

    aligned_colmap_output = dense_colmap_output / 'sparse_txt'
    inliers = find_all_inliers(
        aligned_colmap_output, rendered_gt_poses
    )
    pose_filename = aligned_colmap_output / 'images.txt'
    camera_poses, rendered_est_poses = parse_colmap_poses(pose_filename)

    object_poses = find_object_poses(
        rendered_gt_poses, rendered_est_poses, inliers
    )

    ## using MVS point cloud
    scene_mesh = o3d.io.read_triangle_mesh(
        str(dense_colmap_output / 'meshed-poisson.ply')
    )
    point_cloud = scene_mesh.sample_points_uniformly(
        number_of_points=SCENE_MESH_SAMPLE_POINTS,
        use_triangle_normal=True
    )

    ## using depth point cloud
    # depth_cloud_file = dataset_root / 'scenes' / scene_name / 'depth_cloud.ply'
    # # point_cloud = get_depth_point_cloud(dataset, scene_id, camera_poses)
    # # # lots of redundant points, downsaple for efficiency
    # # point_cloud = point_cloud.voxel_down_sample(voxel_size=DEPTH_CLOUD_RESOLUTION)
    # # o3d.io.write_point_cloud(str(depth_cloud_file), point_cloud)
    # point_cloud = o3d.io.read_point_cloud(str(depth_cloud_file))
    # point_cloud.estimate_normals(
    #     search_param=o3d.geometry.KDTreeSearchParamHybrid(radius=0.1, max_nn=30)
    # )
    # point_cloud = point_cloud.crop(get_workspace_bounding_box())


    # object_poses, _ = icp_refine_object_poses(
    #     dataset, object_poses, point_cloud,
    #     ICP_ALIGNMENT_THRESHOLD, use_normal=True
    # )

    out_scene_dir=Path(dataset.root)/ 'scenes'/ scene_name_list[scene_id]
    # np.save(str(out_scene_dir/'alignment_transform.npy'), alignment_transform)
    # np.save(str(out_scene_dir/'camera_poses.npy'), camera_poses)
    # np.save(str(out_scene_dir/'object_poses.npy'), object_poses)

    # point_cloud = o3d.io.read_point_cloud(str(dense_colmap_output / 'fused.ply'))
    viz_objects(
        dataset, scene_id, object_poses, resize_factor,
        camera_poses, rendered_est_poses, point_cloud
    )


def align_objects_sim_transforms(colmap_output, rendered_gt_poses):
    """find sim transforms for each object to align with gt poses"""
    aligned_colmap_outputs = {}
    max_inliers = 0
    min_error = np.inf
    best_object = None
    best_inliers = []

    for scene_object_name in rendered_gt_poses.keys():
        logging.info('aligning using {}'.format(scene_object_name))

        aligned_colmap_outputs[scene_object_name], alignment_error \
            = align_object_sim_transform(
                scene_object_name, rendered_gt_poses[scene_object_name],
                colmap_output
            )
        inliers = find_all_inliers(
            aligned_colmap_outputs[scene_object_name], rendered_gt_poses
        )
        num_inliers = np.sum([len(i) for i in inliers.values()])
        if num_inliers >= max_inliers and alignment_error < min_error:
            max_inliers = num_inliers
            min_error = alignment_error
            best_object = scene_object_name
            best_inliers = inliers

    for scene_object_name, output in aligned_colmap_outputs.items():
        if scene_object_name != best_object:
            shutil.rmtree(output)

    logging.info('best object for alignment: {}'.format(best_object))
    return aligned_colmap_outputs[best_object], best_inliers


def align_object_sim_transform(scene_object_name, poses, colmap_output):
        timestamp = str(int(time.time()))
        gt_poses_file = Path('/tmp/gt_poses_{}_{}.txt'.format(
            scene_object_name, timestamp
        ))

        with open(str(gt_poses_file), 'w') as f:
            for image_id, pose in poses.items():
                f.write('%s_%.4d.png %f %f %f\n' % (
                    scene_object_name, image_id,
                    *(pose[:3, -1].transpose().tolist())
                ))

        # sim transform alignment
        aligned_colmap_output_bin = Path('/tmp/colmap_output_bin_{}_{}'.format(
            scene_object_name, timestamp
        ))
        aligned_colmap_output_bin.mkdir()
        cmd = [
            'colmap', 'model_aligner',
            '--input_path', str(colmap_output),
            '--output_path', str(aligned_colmap_output_bin),
            '--ref_images_path', str(gt_poses_file),
            '--robust_alignment', str(1),
            '--robust_alignment_max_error', str(ROBUST_ALIGNMENT_MAX_ERR),
            '--alignment_type', 'ECEF'
        ]
        pipe = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        output, _ = pipe.communicate()
        if pipe.returncode != 0:
            logging.warning('Problem with model aligner, exiting.')
            exit(pipe.returncode)

        alignment_error = get_alignment_error(output)

        # convert output to txt (TODO: work with binaries directly)
        aligned_colmap_output_txt = Path('/tmp/colmap_output_txt_{}_{}'.format(
            scene_object_name, timestamp
        ))
        aligned_colmap_output_txt.mkdir()
        cmd = [
            'colmap', 'model_converter',
            '--input_path', str(aligned_colmap_output_bin),
            '--output_path', str(aligned_colmap_output_txt),
            '--output_type', 'TXT'
        ]
        ret = subprocess.call(cmd)
        if ret != 0:
            logging.warning('Problem with model converter, exiting.')
            exit(ret)

        shutil.rmtree(aligned_colmap_output_bin)
        gt_poses_file.unlink()

        return aligned_colmap_output_txt, alignment_error


def get_alignment_error(output):
    error_match = re.match(r'.*(-?\d+\.\d+) \(median\).*', str(output))
    if not error_match:
        logging.error('alignment error missing')
        exit(1)
    alignment_error = error_match.group(1)
    logging.info('alignment error: {}'.format(alignment_error))
    return float(alignment_error)


def find_all_inliers(aligned_colmap_output, rendered_gt_poses):
    """find inliers based on aligned colmap output"""
    inliers = {}
    pose_filename = aligned_colmap_output / 'images.txt'
    _, rendered_est_poses = parse_colmap_poses(pose_filename)

    for scene_object_name in rendered_est_poses.keys():
        inliers[scene_object_name] = find_inliers(
            rendered_gt_poses[scene_object_name],
            rendered_est_poses[scene_object_name]
        )
        print('%s inliers %d/%d' % (
            scene_object_name,
            len(inliers[scene_object_name]),
            len(rendered_est_poses[scene_object_name])
        ))
    return inliers


def do_dense_reconstruction(dataset, scene_name, sparse_model, out_dir):
    timestamp = str(int(time.time()))
    #dense_model = Path('/tmp/dense_model_{}'.format(timestamp))

    image_path = dataset.root / 'scenes' / scene_name / 'rgb_undistort'
    # image_path = dataset.root / 'scenes' / scene_name / 'all_images'

    cmd = [
        'colmap', 'image_undistorter',
        '--input_path', str(sparse_model),
        '--output_path', str(out_dir),
        '--image_path', str(image_path)
    ]
    ret = subprocess.call(cmd)
    if ret != 0:
        logging.warning('Problem with image undistorter, exiting.')
        exit(ret)

    cmd = [
        'colmap', 'patch_match_stereo',
        '--workspace_path', str(out_dir),
        '--PatchMatchStereo.allow_missing_files', str(1),
        '--PatchMatchStereo.filter', 'true',
        '--PatchMatchStereo.geom_consistency', 'true',
        '--PatchMatchStereo.depth_max', str(7),
        '--PatchMatchStereo.depth_min', str(0),
        '--PatchMatchStereo.gpu_index', str(0),
        '--PatchMatchStereo.cache_size', str(64),
    ]
    ret = subprocess.call(cmd)
    if ret != 0:
        logging.warning('Problem with patch match stereo, exiting.')
        exit(ret)

    cmd = [
        'colmap', 'stereo_fusion',
        '--workspace_path', str(out_dir),
        '--StereoFusion.use_cache', str(1),
        '--StereoFusion.cache_size', str(64),
        '--output_path', str(out_dir / 'fused.ply'),
    ]
    ret = subprocess.call(cmd)
    if ret != 0:
        logging.warning('Problem with stereo fusion, exiting.')
        exit(ret)

    # convert output to txt (TODO: work with binaries directly)
    new_sparse_model_bin = out_dir / 'sparse'
    new_sparse_model_txt = out_dir / 'sparse_txt'
    new_sparse_model_txt.mkdir(exist_ok=True)
    cmd = [
        'colmap', 'model_converter',
        '--input_path', str(new_sparse_model_bin),
        '--output_path', str(new_sparse_model_txt),
        '--output_type', 'TXT'
    ]
    ret = subprocess.call(cmd)
    if ret != 0:
        logging.warning('Problem with model converter, exiting.')
        exit(ret)

    cmd = [
        'colmap', 'poisson_mesher',
        '--input_path',  str(out_dir / 'fused.ply'),
        '--output_path',  str(out_dir / 'meshed-poisson.ply')
    ]
    ret = subprocess.call(cmd)
    if ret != 0:
        logging.warning('Problem with poisson mesher, exiting.')
        exit(ret)

    cmd = [
        'colmap', 'delaunay_mesher',
        '--input_path',  str(out_dir),
        '--output_path',  str(out_dir / 'meshed-delaunay.ply'),
        '--input_type', 'dense'
    ]
    ret = subprocess.call(cmd)
    if ret != 0:
        logging.warning('Problem with delaunay mesher, exiting.')
        exit(ret)

    return out_dir


def icp_refine_object_poses(
        dataset, object_poses, scene_pcd, alignment_threshold, use_normal=True
):
    new_object_poses = copy.deepcopy(object_poses)
    pose_quality = {i: None for i in object_poses.keys()}


    for scene_object_name, T_world_object in object_poses.items():
        if T_world_object is None:
            continue
        mesh_path = dataset.root / 'models' / scene_object_name / \
                        'visual.ply'
        o3d_mesh = o3d.io.read_triangle_mesh(str(mesh_path), True)
        object_pcd = o3d_mesh.sample_points_uniformly(
            number_of_points=OBJECT_MESH_SAMPLE_POINTS,
            use_triangle_normal=True
        )

        try:
            registration = o3d.registration
        except AttributeError:
            registration = o3d.pipelines.registration

        if use_normal:
            kwargs = {
                'estimation_method': \
                    registration.TransformationEstimationPointToPlane()
            }
            # object_pcd.estimate_normals(
            #     search_param=o3d.geometry.KDTreeSearchParamHybrid(radius=0.1, max_nn=30)
            # )
        else:
            kwargs = {
                'estimation_method': \
                    registration.TransformationEstimationPointToPoint()
            }


        reg_p2l = registration.registration_icp(
            object_pcd, scene_pcd, alignment_threshold, T_world_object,
            criteria=registration.ICPConvergenceCriteria(
                relative_fitness=1e-6,
                relative_rmse=1e-6,
                max_iteration=1000
            ),
            **kwargs

            # TODO: load color in object model and do colored ICP
            # registration.TransformationEstimationForColoredICP(),
        )
        new_object_poses[scene_object_name] = reg_p2l.transformation
        pose_quality[scene_object_name] = {
            'fitness': reg_p2l.fitness,
            'inlier_rmse': reg_p2l.inlier_rmse,
            'correspondences': np.asarray(reg_p2l.correspondence_set).shape[0]
        }

        # print(reg_p2l)
        # print("Transformation is:")
        # print(reg_p2l.transformation)

    return new_object_poses, pose_quality


def find_object_poses(rendered_gt_poses, rendered_est_poses, inliers):
    object_poses = {}
    for scene_object_name in rendered_est_poses.keys():
        if len(inliers[scene_object_name]) == 0:
            object_poses[scene_object_name] = None
            print(scene_object_name, 'pose cannot be estimated')
            continue
        # just choose the first inlier to find pose
        # TODO: can we do better?
        image_id = inliers[scene_object_name][0]
        T_obj_cam = rendered_gt_poses[scene_object_name][image_id]
        T_world_cam = rendered_est_poses[scene_object_name][image_id]
        T_world_obj = np.dot(T_world_cam, np.linalg.inv(T_obj_cam))
        object_poses[scene_object_name] = T_world_obj
    return object_poses


def viz_objects(
    dataset, scene_id, object_poses, resize_factor,
    camera_poses, rendered_est_poses, point_cloud
):
    geometry_list = []
    for scene_object_name, T_world_object in object_poses.items():
        if T_world_object is None:
            continue
        mesh_path = dataset.root / 'models' / scene_object_name / \
                        'visual.ply'
        o3d_mesh = o3d.io.read_triangle_mesh(str(mesh_path))
        o3d_mesh_world = copy.deepcopy(o3d_mesh).transform(T_world_object)
        ## for debug: save object mesh in world frame
        # o3d.io.write_triangle_mesh(
        #     '/tmp/{}_{}.ply'.format(scene_object_name, scene_id),
        #     o3d_mesh_world
        # )
        geometry_list.append(
            o3d.geometry.LineSet.create_from_triangle_mesh(o3d_mesh_world)
        )

        # show rendered cams
        # for T_world_cam in rendered_est_poses[scene_object_name]:
        #     if T_world_cam is None:
        #         continue
        #     geometry_list.append(
        #         o3d.geometry.TriangleMesh.create_coordinate_frame(0.05) \
        #                     .transform(T_world_cam)
        #     )

    # point_cloud = get_depth_point_cloud(dataset, scene_id, camera_poses)

    geometry_list.append(point_cloud)

    # for T_world_cam in camera_poses.values():
    #     geometry_list.append(
    #         o3d.geometry.TriangleMesh.create_coordinate_frame(0.05) \
    #                     .transform(T_world_cam)
    #     )

    # origin frame
    geometry_list.append(o3d.geometry.TriangleMesh.create_coordinate_frame(0.1))

    # o3d.visualization.RenderOption.point_size = 1e-6
    # o3d.visualization.draw_geometries(geometry_list)
    # return

    seg_masks = render_object_masks(
        dataset, scene_id, camera_poses, object_poses, resize_factor
    )
    viz_seg_masks(dataset, scene_id, seg_masks)


def get_depth_point_cloud(dataset, scene_id, camera_poses):
    scene_name_list = dataset.load_scene_name_list()
    scene_name = scene_name_list[scene_id]
    image_dir = dataset.root / 'scenes' / scene_name / 'rgb_undistort'

    depth_cloud = o3d.geometry.PointCloud()
    depth_cloud.colors = depth_cloud.points

    for image_path in image_dir.iterdir():
        if not image_path.name.endswith('.png') \
                and not image_path.name.endswith('.jpg'):
            continue
        image_id = int(image_path.stem)

        ## using COLMAP poses
        if image_id in camera_poses:
            pcd = dataset.load_point_cloud(scene_id=scene_id, image_id=image_id)
            pcd.transform(camera_poses[image_id])
            depth_cloud.points.extend(pcd.points)
            depth_cloud.colors.extend(pcd.colors)

        ## using forward kinematic poses
        # pcd = dataset.load_point_cloud(scene_id=scene_id, image_id=image_id)
        # camera_pose = dataset.load_camera_pose(scene_id, image_id)
        # pcd.transform(camera_pose)
        # geometry_list.append(pcd)

    return depth_cloud


def viz_seg_masks(dataset, scene_id, seg_masks):
    timestamp = str(int(time.time()))
    scene_name_list = dataset.load_scene_name_list()
    scene_name = scene_name_list[scene_id]
    seg_masks_path = Path(dataset.root) / 'scenes' / scene_name \
                        / 'seg_masks_{}'.format(timestamp)
    seg_masks_path.mkdir(exist_ok=True)

    for image_id, seg_mask in tqdm.tqdm(seg_masks.items()):
        rgb = dataset.load_raw_image(
            scene_id=scene_id, image_id=image_id, order = 'BGR'
        )
        # unique_ids = np.unique(seg_mask)

        # seg_mask_colorized = cv2.applyColorMap(
        #     (seg_mask.astype(np.float32) / max(unique_ids) * 255).astype(np.uint8),
        #     cv2.COLORMAP_JET
        # )
        # non_zero_mask = np.repeat(np.expand_dims(seg_mask != 0, -1), 3, -1)
        # seg_mask_colorized *= non_zero_mask

        seg_mask_colorized = np.zeros(rgb.shape, dtype=np.uint8)
        # seg_mask_colorized[:, :, 2] = ((seg_mask != 0) * 255)

        rgb *= np.stack((seg_mask == 0, seg_mask == 0, seg_mask == 0), axis=-1)

        alpha = 0.4
        cv2.imwrite(
            str(seg_masks_path / '{:05d}.png'.format(image_id)),
            cv2.addWeighted(rgb, 1.0, seg_mask_colorized, alpha, 0)
        )


def render_object_masks(
        dataset, scene_id, camera_poses, object_poses, resize_factor
):
    seg_masks = {}
    # object_meshes_world = {
    #     scene_object_name: \
    #         o3d.io.read_triangle_mesh(str(
    #             dataset.root / 'models' / scene_object_name / 'visual.ply'
    #         ))
    #     for scene_object_name in object_poses.keys()
    # }

    # use OpenGL convention (East-Up-South) instead of East-Down-North
    rotation_correction = np.asarray([
        [1, 0, 0], [0, -1, 0], [0, 0, -1]
    ])

    camera_param = dataset.load_camera_param(scene_id)
    camera_param = resize_camera_parameters(camera_param, resize_factor)
    print(
        f'width: {camera_param.intrinsic.width}, '
        f'height: {camera_param.intrinsic.height}, '
        f'f: {camera_param.intrinsic.get_focal_length()}, '
        f'c: {camera_param.intrinsic.get_principal_point()}, '
    )
    ## open3d 0.9 does not support rendering when width != height
    # vis = o3d.visualization.Visualizer()
    # vis.create_window(
    #     width=camera_param.intrinsic.width,
    #     height=camera_param.intrinsic.height,
    #     visible=True
    # )

    fx, fy = camera_param.intrinsic.get_focal_length()
    cx, cy = camera_param.intrinsic.get_principal_point()
    cam = pyrender.IntrinsicsCamera(fx=fx, fy=fy, cx=cx, cy=cy)

    pyrender_meshes = {}
    for scene_object_name, T_world_object  in object_poses.items():
        mesh_path = dataset.root / 'models' / scene_object_name / \
                        'visual.ply'

        # draw wireframe for visualization
        o3d_mesh = o3d.io.read_triangle_mesh(str(mesh_path))
        o3d_mesh = o3d_mesh.simplify_quadric_decimation(1000)
        mesh_path = '/tmp/mesh_tmp.obj'
        o3d.io.write_triangle_mesh(mesh_path, o3d_mesh)

        trimesh_mesh = trimesh.load(mesh_path)
        trimesh_mesh.apply_transform(T_world_object)
        pyrender_meshes[scene_object_name] \
            = pyrender.Mesh.from_trimesh(trimesh_mesh)

    logging.info('rendering segmentation masks')

    scene = pyrender.Scene(
        ambient_light=np.array([0.35, 0.35, 0.35, 1.0])
    )

    for scene_object_name, T_world_object in object_poses.items():
        if T_world_object is None:
            continue
        scene.add(pyrender_meshes[scene_object_name])

    cam_node = scene.add(cam)
    r = pyrender.OffscreenRenderer(
        viewport_width=camera_param.intrinsic.width,
        viewport_height=camera_param.intrinsic.height
    )

    for image_id, T_world_cam in tqdm.tqdm(camera_poses.items()):
        geometry_list = []
            ## open3d 0.9 does not support rendering when width != height
            # o3d_mesh_world = copy.deepcopy(
            #     object_meshes_world[scene_object_name]
            # ).transform(T_world_object)
            # o3d_mesh_cam = o3d_mesh_world.transform(np.linalg.inv(T_world_cam))
            # obj_mask_id = (dataset.object_id_dict[scene_object_name]) / 255.0
            # o3d_mesh_cam.texture.clear()
            # o3d_mesh_cam = o3d_mesh_cam.paint_uniform_color(
            #     [obj_mask_id, obj_mask_id, obj_mask_id]
            # )
            # geometry_list.append(o3d_mesh_cam)

        cam_pose = T_world_cam.copy()
        cam_pose[:3, :3] = np.dot(cam_pose[:3, :3], rotation_correction)
        scene.set_pose(cam_node, cam_pose)

        render_flags = pyrender.constants.RenderFlags.ALL_WIREFRAME | pyrender.constants.RenderFlags.RGBA
        rendered_color, rendered_depth = r.render(scene, flags=render_flags)
        seg_mask = rendered_depth > 0

        ## open3d 0.9 does not support rendering when width != height
        # for g in geometry_list:
        #     vis.add_geometry(g)

        # ctr = vis.get_view_control()
        # camera_param.extrinsic = np.eye(4)
        # ctr.convert_from_pinhole_camera_parameters(camera_param, arbitrary=True)
        # vis.poll_events()
        # tmp_file = tempfile.mkstemp('.png')
        # vis.capture_screen_image(tmp_file[1], do_render=True)
        # seg_mask = cv2.imread(tmp_file[1])[:,:,0]
        # Path(tmp_file[1]).unlink()

        # valid_mask = np.zeros(
        #     (camera_param.intrinsic.height, camera_param.intrinsic.width),
        #     dtype=bool
        # )
        # for scene_object_name in object_poses.keys():
        #     obj_id = dataset.object_id_dict[scene_object_name]
        #     valid_mask = valid_mask | (seg_mask == (obj_id))
        # seg_mask[~valid_mask] = 0

        # for g in geometry_list:
        #     vis.remove_geometry(g)

        seg_masks[image_id] = seg_mask

    r.delete()

    # vis.destroy_window()
    # vis.close()

    return seg_masks


def find_inliers(gt_poses, est_poses):
    """
    gt_poses: T_obj_cam (wrt object local frame)
    est_poses: T_world_cam (wrt global colmap frame)
    """
    curr_inliers = []

    T_obj_world = None
    for image_id in est_poses.keys():
        T_obj_i = gt_poses[image_id]
        T_world_i = est_poses[image_id]
        if T_world_i is None:
            continue

        if T_obj_world is None:
            # find transform from just one image
            T_obj_world = np.dot(T_obj_i, np.linalg.inv(T_world_i))
            # first one is always an inlier
            curr_inliers.append(image_id)
            continue

        T_obj_i_recons = np.dot(T_obj_world, T_world_i)

        r_err = get_rotation_error(T_obj_i[:3, :3], T_obj_i_recons[:3, :3])
        if np.any(np.abs(r_err) > ROTATION_ERR_THRESH):
            continue

        t_err = T_obj_i[:3, -1] - T_obj_i_recons[:3, -1]
        if np.any(np.abs(t_err) > TRANSLATION_ERR_THRESH):
            continue
        curr_inliers.append(image_id)
    return curr_inliers


def get_rotation_error(R_world1_obj, R_world2_obj):
    R_err = np.dot(R_world1_obj, R_world2_obj.transpose())
    r_err = Rotation.from_matrix(R_err).as_euler('xyz', degrees=True)
    return r_err


def get_gt_object_poses(dataset, scene_name_list, scene_id, scene_object_name):
    rendered_pose_dir = dataset.root / 'scenes' / scene_name_list[scene_id] \
                            / 'rendered' / scene_object_name / 'camera_pose'
    poses = {}
    for pose_path in rendered_pose_dir.iterdir():
        if not pose_path.name.endswith('.npy'):
            continue
        image_id = int(pose_path.stem)
        T = np.load(str(pose_path))
        poses[image_id] = T

    return poses


def parse_colmap_pose(qw, qx, qy, qz, tx, ty, tz):
    R = Rotation.from_quat([float(qx), float(qy), float(qz), float(qw)]) \
                .as_matrix()
    # the pose in T_camera_world, we need T_world_camera (wrt world)
    t = np.asarray([float(tx), float(ty), float(tz)]).transpose()
    T = np.eye(4)
    T[:3, :3] = R.transpose()
    T[:3, -1] = np.dot(-R.transpose(), t)
    return T


def parse_colmap_poses(pose_filename):
    with open(pose_filename) as f:
        lines = f.readlines()

    camera_poses = {}
    object_poses = {}
    for line in lines[4::2]:
        _, qw, qx, qy, qz, tx, ty, tz, _, image = line.strip().split(' ')
        T = parse_colmap_pose(qw, qx, qy, qz, tx, ty, tz)

        image_id = int(image.split('.')[0].split('_')[-1])
        # if len(image) == 8:
        #     camera_poses[image_id] = T
        # else:
        if True:
            object_name = '_'.join(image.split('.')[0].split('_')[0:-1])
            if object_name not in object_poses:
                object_poses[object_name] = {}
            object_poses[object_name][image_id] = T
    camera_poses = copy.deepcopy(object_poses['color'])
    del object_poses['color']
    return camera_poses, object_poses


def parse_colmap_keypoints(pose_filename):
    with open(pose_filename) as f:
        lines = f.readlines()

    image_ids = []
    object_names = []
    for line in lines[4::2]:
        image = line.strip().split(' ')[-1]
        image_id = int(image.split('.')[0].split('_')[-1])
        object_name = '_'.join(image.split('.')[0].split('_')[0:-1])
        image_ids.append(image_id)
        object_names.append(object_name)

    object_keypoints = {}
    for image_idx, line in enumerate(lines[5::2]):
        kps_pt3dids = line.strip().split(' ')
        kps = np.asarray([float(i) for i in kps_pt3dids]).reshape((-1, 3))[:, :2]
        pt3d_ids = [int(i) for i in kps_pt3dids[2::3]]

        image_id = image_ids[image_idx]
        object_name = object_names[image_idx]

        if object_name not in object_keypoints:
            object_keypoints[object_name] = {}

        object_keypoints[object_name][image_id] = {i:j for i, j in zip(pt3d_ids, kps)}

    camera_keypoints = copy.deepcopy(object_keypoints['color'])
    del object_keypoints['color']

    return camera_keypoints, object_keypoints


def parse_colmap_points(points_filename):
    with open(points_filename) as f:
        lines = f.readlines()

    points = np.zeros((len(lines) - 3, 3))
    colors = np.zeros((len(lines) - 3, 3))

    for line_idx, line in enumerate(lines[3:]):
        row = line.split(' ')
        x, y, z = float(row[1]), float(row[2]), float(row[3])
        r, g, b = float(row[4]), float(row[5]), float(row[6])
        r, g, b = r/255., g/255., b/255.
        points[line_idx, :] = np.asarray([x, y, z])
        colors[line_idx, :] = np.asarray([r, g, b])

    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(points)
    pcd.colors = o3d.utility.Vector3dVector(colors)
    return pcd


def get_kinematic_cam_poses(dataset, scene_name):
    kinematic_cam_poses = {}
    cam_pose_dir = Path(dataset.root) / 'scenes' / scene_name / 'camera_pose'
    for pose_path in cam_pose_dir.iterdir():
        if not pose_path.name.endswith('.npy'):
            continue
        image_id = int(pose_path.stem)
        kinematic_cam_poses[image_id] = np.load(str(pose_path))

    return kinematic_cam_poses


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--dataset-root', type=Path,
        default='/root/3d_dataset'
    )
    parser.add_argument(
        '--scene-name', type=str,
        default='scene2'
    )
    parser.add_argument('--resize-factor', type=int, default=2)
    args = parser.parse_args()
    main(**args.__dict__)
