import sys
sys.path.append('dataset_toolkit')
sys.path.append('Hierarchical-Localization')
import argparse
from pathlib import Path
import numpy as np
from scipy.spatial.transform import Rotation
import subprocess
import multiprocessing
import logging
import pprint
import itertools

from hloc import extract_features, match_features
from hloc import reconstruction, visualization, triangulation
from hloc.utils.database import COLMAPDatabase
from hloc.utils.read_write_model import read_cameras_binary

from toc_dataset_toolkit import TOC_Dataset
from toc_dataset_toolkit.utils.logging import set_log_level


def main(
    images, outputs, skip_geometric_verification,
    extract_feat=True, feature_conf='superpoint_inloc'
):
    sfm_pairs = outputs / 'pairs-exhaustive.txt'  # exhaustive matching
    sfm_dir = outputs / 'sfm_superpoint+superglue'

    feature_conf = extract_features.confs[feature_conf]
    matcher_conf = match_features.confs['superglue']

    features = feature_conf['output']
    feature_file = f"{features}.h5"
    match_file = f"{features}_{matcher_conf['output']}_{sfm_pairs.stem}.h5"

    outputs.mkdir(exist_ok=True, parents=True)
    # generate_pairs(images, sfm_pairs)
    if extract_feat:
        extract_features.main(feature_conf, images, outputs)
        match_features.main(matcher_conf, sfm_pairs, features, outputs, exhaustive=True) # False)
    create_colmap_database(
        sfm_dir, images, sfm_pairs,
        outputs / feature_file, outputs / match_file,
        skip_geometric_verification
    )


def generate_pairs(images, sfm_pairs):
    """finds pairs of images between which feature matching is done
    avoid doing feature matching between two different rendered object"""
    cameras = []
    objects = {}

    for image in list(images.iterdir()):
        if (not str(image).endswith('.png')) \
                and (not str(image).endswith('.jpg')):
            continue
        if len(image.stem) == 4:
            cameras.append(image.name)
        else:
            object_name = '_'.join(image.stem.split('_')[0:-1])
            if object_name not in objects:
                objects[object_name] = []
            objects[object_name].append(image.name)

    pairs = []
    pairs.extend(itertools.combinations(cameras, 2))
    for object_name, object_images in objects.items():
        pairs.extend(itertools.combinations(object_images, 2))
        pairs.extend(itertools.product(object_images, cameras))

    with open(str(sfm_pairs), 'w') as f:
        f.write('\n'.join([' '.join(pair) for pair in pairs]))


def import_images(sfm_dir, image_dir, database):
    images = list(image_dir.iterdir())
    if len(images) == 0:
        raise IOError(f'No images found in {image_dir}.')

    scene_dir = image_dir.parents[0]
    scene_name = scene_dir.stem
    toc_root = scene_dir.parents[1]
    toc_dataset = TOC_Dataset(root=toc_root)
    scene_name_list = toc_dataset.load_scene_name_list()
    scene_id = scene_name_list.index(scene_name)
    camera_param = toc_dataset.load_camera_param(scene_id)

    # only one camera (params) for all images
    fx, fy = camera_param.intrinsic.get_focal_length()
    cx, cy = camera_param.intrinsic.get_principal_point()
    # simple pinhole camera model (1)
    camera_model = 1
    database.add_camera(
        camera_model,
        camera_param.intrinsic.width, camera_param.intrinsic.height,
        [fx, fy, cx, cy], prior_focal_length=True
    )

    add_image_poses(
        database, images, scene_dir, scene_name, scene_id, toc_dataset
    )


def add_image_poses(
    database, images, scene_dir, scene_name, scene_id, toc_dataset
):
    object_pose_dict = toc_dataset.load_object_pose_dict(scene_id)

    for image_path in images:
        image_stem = image_path.stem
        try:
            image_id = int(image_stem.split('_')[-1])
        except ValueError:
            logging.error('{} not a valid filename'.format(image_stem))
            continue

        if len(image_stem) == 4:
            # T_world_cam = toc_dataset.load_camera_pose(scene_id, image_id)
            # assume we don't know camera pose
            T_world_cam = np.eye(4)
        else:
            # assume we don't know object pose
            T_world_cam = np.eye(4)
            # object_name = '_'.join(image_stem.split('_')[0:-1])
            # cam_pose_path = scene_dir / 'rendered' / object_name \
            #                     / 'camera_pose' / ('%.04d.npy' % image_id)
            # T_obj_cam = np.load(str(cam_pose_path))
            # T_world_obj = object_pose_dict[object_name]
            # T_world_cam = np.dot(T_world_obj, T_obj_cam)

        T_cam_world = np.linalg.inv(T_world_cam)
        qx, qy, qz, qw = Rotation.from_matrix(T_cam_world[:3, :3]).as_quat()
        database.add_image(
            image_path.name, 1, np.asarray([qw, qx, qy, qz]),
            T_cam_world[:3, -1].transpose()
        )


def get_image_ids(db):
    images = {}
    for name, image_id in db.execute("SELECT name, image_id FROM images;"):
        images[name] = image_id
    return images


def create_empty_db(database_path):
    db = COLMAPDatabase.connect(database_path)
    db.create_tables()
    db.commit()
    return db


def run_reconstruction(model_path, database_path, image_dir,
                       min_num_matches=None):
    logging.info('Running the 3D reconstruction...')
    model_path.mkdir(exist_ok=True)

    cmd = [
        'colmap', 'mapper',
        '--database_path', str(database_path),
        '--image_path', str(image_dir),
        '--output_path', str(model_path),
        '--Mapper.num_threads', str(min(multiprocessing.cpu_count(), 16))]
    if min_num_matches:
        cmd += ['--Mapper.min_num_matches', str(min_num_matches)]
    logging.info(' '.join(cmd))
    ret = subprocess.call(cmd)
    if ret != 0:
        logging.warning('Problem with mapper, exiting.')
        exit(ret)

    models = list(model_path.iterdir())
    if len(models) == 0:
        logging.error('Could not reconstruct any model!')
        return False
    logging.info(f'Reconstructed {len(models)} models.')

    largest_model = None
    largest_model_num_images = 0
    for model in models:
        num_images = len(read_cameras_binary(str(model / 'cameras.bin')))
        if num_images > largest_model_num_images:
            largest_model = model
            largest_model_num_images = num_images
    assert largest_model_num_images > 0
    logging.info(f'Largest model is #{largest_model.name} '
                 'with {largest_model_num_images} images.')

    stats_raw = subprocess.check_output(
        ['colmap', 'model_analyzer',
         '--path', str(largest_model)])
    stats_raw = stats_raw.decode().split("\n")
    stats = dict()
    for stat in stats_raw:
        if stat.startswith("Registered images"):
            stats['num_reg_images'] = int(stat.split()[-1])
        elif stat.startswith("Points"):
            stats['num_sparse_points'] = int(stat.split()[-1])
        elif stat.startswith("Observations"):
            stats['num_observations'] = int(stat.split()[-1])
        elif stat.startswith("Mean track length"):
            stats['mean_track_length'] = float(stat.split()[-1])
        elif stat.startswith("Mean observations per image"):
            stats['num_observations_per_image'] = float(stat.split()[-1])
        elif stat.startswith("Mean reprojection error"):
            stats['mean_reproj_error'] = float(stat.split()[-1][:-2])

    return stats


def create_colmap_database(
    sfm_dir, images, pairs, features, matches, skip_geometric_verification
):
    assert features.exists(), features
    assert pairs.exists(), pairs
    assert matches.exists(), matches

    sfm_dir.mkdir(parents=True, exist_ok=True)
    models = sfm_dir / 'models'
    models.mkdir(exist_ok=True)

    database_path = sfm_dir / 'database.db'
    if database_path.exists():
        # mission_ok only available from 3.8
        database_path.unlink() # missing_ok=True)

    database = create_empty_db(database_path)
    import_images(sfm_dir, images, database)
    database.commit()
    image_ids = get_image_ids(database)
    database.close()

    triangulation.import_features(image_ids, database_path, features)
    triangulation.import_matches(
        image_ids, database_path, pairs, matches,
        skip_geometric_verification=skip_geometric_verification,
        min_match_score=None
    )

    if not skip_geometric_verification:
        triangulation.geometric_verification('colmap', database_path, pairs)
    stats = run_reconstruction(
        models, database_path, images, min_num_matches=None)
    stats['num_input_images'] = len(image_ids)
    print(f'Statistics:\n{pprint.pformat(stats)}')


if __name__ == '__main__':
    set_log_level('WARNING')
    parser = argparse.ArgumentParser()
    parser.add_argument('--images', type=Path, required=True)
    parser.add_argument('--outputs', type=Path, required=True)
    parser.add_argument('--skip_geometric_verification', action='store_true')
    parser.add_argument('--feature_conf', type=str, default='superpoint_inloc')
    args = parser.parse_args()
    main(**args.__dict__)
