#!/usr/bin/env python
import sys
import os
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'dataset_toolkit'))


from toc_dataset_toolkit import TOC_Dataset
from toc_dataset_toolkit.utils.logging import set_log_level

import argparse
from pathlib import Path
import tqdm
import random
import itertools
import json
import numpy as np
from scipy.special import comb

IMAGE_TUPLE_SIZE = 3
NUM_IMAGE_TUPLES = 500


def main(dataset_root):
    set_log_level('WARNING')

    scene_name_list = [i.name for i in dataset_root.iterdir() if i.is_dir()]
    scene_name_list.sort()

    scene_image_ids = {}

    for scene_name in scene_name_list:
        scene_dir = dataset_root / scene_name
        image_dir = scene_dir / 'meshmvs_training_images'
        camera_poses_file = scene_dir / 'meshmvs_camera_poses.npy'
        if not camera_poses_file.is_file():
            print(f'{camera_poses_file} not found')
            continue
        camera_poses = np.load(str(camera_poses_file), allow_pickle=True)[()]

        scene_images = []

        for image_id in camera_poses.keys():
            if not (image_dir / f'color_{image_id}.png').is_file() \
                    or not (image_dir / f'depth_{image_id}.png').is_file():
                continue
            scene_images.append(image_id)

        if len(scene_images):
            scene_image_ids[scene_name] = scene_images
            scene_image_ids[scene_name].sort()

    # group based on category
    category_scenes_map, scene_category_map = get_category_dict(dataset_root)
    category_image_ids = {category: {} for category in category_scenes_map.keys()}

    for scene_name, image_ids in scene_image_ids.items():
        category = scene_category_map.get(scene_name, None)
        if category is None:
            print(f'{scene_name} not in scene_category_map')
            continue
        category_image_ids[category][scene_name] = image_ids

    train_scenes = []
    test_scenes = []
    val_scenes = []

    for category, category_scenes in category_image_ids.items():
        scene_names = list(category_scenes.keys())
        random.shuffle(scene_names)
        train_count = int(np.ceil(len(scene_names) * 0.8))
        test_count = int(np.floor(len(scene_names) * 0.1))
        val_count = len(scene_names) - train_count - test_count

        train_scenes.extend(scene_names[:train_count])
        test_scenes.extend(scene_names[train_count:(train_count+test_count)])
        val_scenes.extend(scene_names[(train_count+test_count):])

    # take n combinations of image tuples
    def sample_images(t_set):
        nonlocal scene_image_ids
        output = []
        for scene in t_set:
            image_ids = scene_image_ids[scene]

            ## rank based combinations, an image can repeat in the tuples
            nCk = comb(len(image_ids), IMAGE_TUPLE_SIZE)

            image_tuples = []
            # ranks in the combination already selected
            selected_ranks = {}
            while len(image_tuples) < NUM_IMAGE_TUPLES:
                rank = min(int(np.random.random() * nCk), nCk - 1)
                # rank starts from 1
                rank += 1
                if selected_ranks.get(rank, False):
                    continue

                selected_ranks[rank] = True
                unranked = ksub_unrank(IMAGE_TUPLE_SIZE, rank)
                selected_ids = sorted([image_ids[int(i)-1] for i in unranked])
                selected_ids = tuple(selected_ids)
                image_tuples.append(selected_ids)

            ## naive combination, no repeat
            # random.shuffle(image_ids)
            # trunc_size = (len(image_ids) // IMAGE_TUPLE_SIZE) * IMAGE_TUPLE_SIZE
            # image_ids = image_ids[:trunc_size]
            # image_ids = np.asarray(image_ids).reshape((-1, IMAGE_TUPLE_SIZE))
            # image_ids = image_ids[:NUM_IMAGE_TUPLES, :]
            # image_tuples = image_ids.tolist()

            output.append({
                'scene': scene,
                'image_ids': image_tuples
            })
        return output

    output_json = {}

    output_json['train'] = sample_images(train_scenes)
    output_json['test'] = sample_images(test_scenes)
    output_json['val'] = sample_images(val_scenes)

    with open('splits_val05.json', 'w') as f:
        json.dump(output_json, f, indent=4)


def get_category_dict(dataset_root):
    categories_file = dataset_root / 'categories.json'
    with open(categories_file, 'r') as f:
        category_dict = json.load(f)

    scene_dict = {}
    for category, scenes in category_dict.items():
        for scene in scenes:
            scene_dict[scene] = category

    return category_dict, scene_dict


def ksub_unrank ( k, rank ):
#*****************************************************************************80
#
## KSUB_UNRANK returns the subset of a given rank.
#  https://people.math.sc.edu/Burkardt/py_src/subset/ksub_unrank.py
#
#  Discussion:
#
#    The routine is given a rank and returns the corresponding subset of K
#    elements of a set of N elements.
#
#    It uses the same ranking that KSUB_NEXT2 uses to generate all the subsets
#    one at a time.
#
#    Note that the value of N itself is not input, nor is it needed.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    25 May 2015
#
#  Author:
#
#    John Burkardt.
#
#  Reference:
#
#    Albert Nijenhuis, Herbert Wilf,
#    Combinatorial Algorithms,
#    Academic Press, 1978, second edition,
#    ISBN 0-12-519260-6.
#
#  Parameters:
#
#    Input, integer K, the number of elements in the subset.
#
#    Input, integer RANK, the rank of the desired subset.
#    There are ( N*(N-1)*...*(N+K-1)) / ( K*(K-1)*...*2*1) such
#    subsets, so RANK must be between 1 and that value.
#
#    Output, integer A(K), K distinct integers in order between
#    1 and N, which define the subset.
#

    a = np.zeros ( k )

    jrank = rank - 1

    for i in range ( k, 0, -1 ):

        ip = i - 1
        iprod = 1

        while ( True ):

            ip = ip + 1

            if ( ip != i ):
                iprod = ( ( ip * iprod ) // ( ip - i ) )

            if ( jrank < iprod ):
                break

        if ( ip != i ):
            iprod = ( ( ( ip - i ) * iprod ) // ip )

        jrank = jrank - iprod
        a[i-1] = ip

    return a


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--dataset-root', type=Path,
        default='/root/3d_dataset'
    )
    args = parser.parse_args()
    main(**args.__dict__)
