#!/usr/bin/env python
import sys
import os
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'dataset_toolkit'))

from toc_dataset_toolkit import TOC_Dataset
from toc_dataset_toolkit.utils.logging import set_log_level

import argparse
from copy import deepcopy
import itertools
from pathlib import Path
import tqdm

import cv2
import numpy as np
import open3d as o3d
import trimesh

from pose_estimation import (
    get_gt_object_poses
)


def main(dataset_root, scene_name, resize_factor):
    set_log_level('WARNING')

    dataset = TOC_Dataset(root=dataset_root)
    scene_name_list = dataset.load_scene_name_list()
    scene_id = scene_name_list.index(scene_name)
    print('scene', scene_id, scene_name)

    scene_dir = dataset_root / 'scenes' / scene_name

    recons_file = scene_dir / 'fastfusion_reconstruction.ply'
    scene_mesh = o3d.io.read_triangle_mesh(str(recons_file))
    scene_pcd = o3d.geometry.PointCloud(scene_mesh.vertices)
    scene_pcd.colors = scene_mesh.vertex_colors

    camera_poses = np.load(
        str(scene_dir / 'camera_poses.npy'), allow_pickle=True
    ).item()
    object_poses = np.load(
        str(scene_dir / 'object_poses.npy'), allow_pickle=True
    ).item()

    object_name = next(iter(object_poses.keys()))
    object_pose = next(iter(object_poses.values()))

    rendered_poses = get_gt_object_poses(
        dataset, scene_name_list, scene_id, object_name
    )
    rendered_poses = [
        i for i in rendered_poses.values()
        if i[2, 3] > 0.1
    ]

    object_mesh_path = dataset_root / 'models' / object_name / 'visual.ply'
    object_mesh = o3d.io.read_triangle_mesh(str(object_mesh_path))
    object_mesh = object_mesh.simplify_quadric_decimation(1000)

    camera_frustum = [
        get_camera_frustum(0.025, 0.05).transform(i)
            .paint_uniform_color([0.75, 0, 0])
        for i in list(camera_poses.values())[::3]
    ]

    o3d.visualization.draw_geometries([scene_pcd,
        *[
            get_camera_frustum(0.025, 0.05).transform(i)
                .paint_uniform_color([0.75, 0, 0])
            for i in list(camera_poses.values())[::3]
        ],
        # *[
        #     get_camera_frustum(0.025, 0.05).transform(i)
        #         .paint_uniform_color([0, 0.75, 0])
        #     for i in rendered_poses[::2]
        # ]
    ])

    o3d.visualization.draw_geometries([
        o3d.geometry.LineSet.create_from_triangle_mesh(
            object_mesh
        ).transform(object_pose),
        object_mesh.transform(object_pose),
        scene_pcd,
        # *camera_frustum,
        # *[o3d.geometry.LineSet.create_from_triangle_mesh(i) for i in camera_frustum],
        o3d.geometry.TriangleMesh
            .create_coordinate_frame(size=0.1, origin=[0, 0, 0]) \
            .transform(object_pose)
    ])


def get_camera_frustum(size, z):
    w = h = size
    vertices = o3d.utility.Vector3dVector(np.asarray([
        [0, 0, 0], [-w, h, z], [w, h, z], [w, -h, z], [-w, -h, z]
    ]))
    faces = o3d.utility.Vector3iVector(np.asarray([
        [2, 1, 3], [3, 1, 4], [0, 3, 4], [0, 2, 3], [0, 1, 2], [0, 4, 1]
    ], dtype=int))
    frustum = o3d.geometry.TriangleMesh(vertices, faces)
    return frustum


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--dataset-root', type=Path, default='/root/3d_dataset'
    )
    parser.add_argument(
        '--scene-name', type=str, default='scene2'
    )
    parser.add_argument('--resize-factor', type=int, default=2)
    args = parser.parse_args()
    main(**args.__dict__)
